# Generated by Django 3.2.19 on 2023-06-10 03:16

import re

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("NEMO", "0045_version_4_5_5"),
        ("NEMO_billing", "0004_version_2_4_0"),
    ]

    operations = [
        migrations.CreateModel(
            name="FundType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The unique name for this item", max_length=200, unique=True)),
                (
                    "display_order",
                    models.IntegerField(
                        help_text="The display order is used to sort these items. The lowest value category is displayed first."
                    ),
                ),
            ],
            options={
                "ordering": ["display_order", "name"],
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ProjectPrepaymentDetail",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "charge_types",
                    models.CharField(
                        help_text="List of charge types allowed",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^\\d+(?:,\\d+)*\\Z"),
                                code="invalid",
                                message="Enter only digits separated by commas.",
                            )
                        ],
                    ),
                ),
                ("balance_last_updated", models.DateField(blank=True)),
                (
                    "only_core_facilities",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Limit which core facilities are allowed for this project. Leave blank to allow them all",
                        to="NEMO_billing.CoreFacility",
                    ),
                ),
                (
                    "configuration",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="invoices.invoiceconfiguration",
                    ),
                ),
                (
                    "project",
                    models.OneToOneField(
                        help_text="The project",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="NEMO.project",
                        verbose_name="Project",
                    ),
                ),
            ],
            options={
                "ordering": ["project"],
            },
        ),
        migrations.CreateModel(
            name="Fund",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("reference", models.CharField(blank=True, max_length=255, null=True)),
                ("amount", models.DecimalField(decimal_places=2, max_digits=14)),
                (
                    "start_month",
                    models.PositiveIntegerField(
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ]
                    ),
                ),
                (
                    "start_year",
                    models.PositiveIntegerField(
                        validators=[
                            django.core.validators.MinValueValidator(1900),
                            django.core.validators.MaxValueValidator(9999),
                        ]
                    ),
                ),
                (
                    "expiration_month",
                    models.PositiveIntegerField(
                        blank=True,
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ],
                        null=True,
                    ),
                ),
                (
                    "expiration_year",
                    models.PositiveIntegerField(
                        blank=True,
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1900),
                            django.core.validators.MaxValueValidator(9999),
                        ],
                    ),
                ),
                ("balance", models.DecimalField(blank=True, decimal_places=2, max_digits=14)),
                (
                    "balance_warning_percent",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Send a warning email when the balance is below this percent.",
                        null=True,
                        validators=[django.core.validators.MaxValueValidator(100)],
                    ),
                ),
                ("balance_warning_sent", models.DateTimeField(blank=True, null=True)),
                ("note", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "fund_type",
                    models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to="prepayments.fundtype"),
                ),
                (
                    "project_prepayment",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="prepayments.projectprepaymentdetail"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from datetime import datetime, timedelta
from typing import List, Tuple

from NEMO.models import (
    Area,
    AreaAccessRecord,
    Consumable,
    ConsumableWithdraw,
    Reservation,
    Tool,
    TrainingSession,
    UsageEvent,
    User,
)

from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.models import CustomCharge
from NEMO_billing.rates.models import Rate, RateCategory, RateType


def create_items() -> Tuple[Tool, Area, Consumable, User]:
    consumable = Consumable.objects.create(name="Tweezers", quantity=10, reusable=True)
    consumable_type = RateType.objects.get(type="CONSUMABLE")
    rate_category = RateCategory.objects.get(name="Academia")
    Rate.objects.create(category=rate_category, type=consumable_type, consumable=consumable, flat=True, amount="20.00")
    tool = Tool.objects.create(name="Tool", _operational=True)
    tool_type = RateType.objects.get(type="TOOL_USAGE")
    Rate.objects.create(category=rate_category, type=tool_type, tool=tool, amount="30.00")
    area = Area.objects.create(name="Area")
    area_type = RateType.objects.get(type="AREA_USAGE")
    Rate.objects.create(category=rate_category, type=area_type, area=area, amount="40.00")
    staff = User.objects.create(username="staff", first_name="Staff", last_name="Staff", is_staff=True, badge_number=2)
    tool_missed_resa_type = RateType.objects.get(type="TOOL_MISSED_RESERVATION")
    Rate.objects.create(category=rate_category, type=tool_missed_resa_type, tool=tool, amount="15.00")
    tool_training_type = RateType.objects.get(type="TOOL_TRAINING_INDIVIDUAL")
    Rate.objects.create(category=rate_category, type=tool_training_type, tool=tool, amount="10.00")
    staff_charge_type = RateType.objects.get(type="STAFF_CHARGE")
    Rate.objects.create(category=rate_category, type=staff_charge_type, amount="25.00")
    return tool, area, consumable, staff


def delete_all_billables():
    UsageEvent.objects.all().delete()
    AreaAccessRecord.objects.all().delete()
    Reservation.objects.all().delete()
    ConsumableWithdraw.objects.all().delete()
    TrainingSession.objects.all().delete()
    CustomCharge.objects.all().delete()


def get_other_billable_types(types: List[BillableItemType]) -> List[BillableItemType]:
    result = []
    for billableItemType in BillableItemType:
        if billableItemType not in types:
            result.append(billableItemType)
    return result


def add_charges_of_type(test_case, types: List[BillableItemType], end: datetime, cap_eligible_custom_charge=True):
    user, project, staff, tool, area, consumable = (
        test_case.user,
        test_case.project,
        test_case.staff,
        test_case.tool,
        test_case.area,
        test_case.consumable,
    )
    start = end - timedelta(days=1)
    if BillableItemType.TOOL_USAGE in types:
        UsageEvent.objects.create(user=user, operator=user, tool=tool, project=project, start=start, end=end)
    if BillableItemType.CONSUMABLE in types:
        ConsumableWithdraw.objects.create(
            customer=user,
            project=project,
            consumable=consumable,
            quantity=1,
            date=end,
            merchant=staff,
        )
    if BillableItemType.STAFF_CHARGE in types:
        UsageEvent.objects.create(
            user=user, operator=staff, remote_work=True, tool=tool, project=project, start=start, end=end
        )
    if BillableItemType.AREA_ACCESS in types:
        AreaAccessRecord.objects.create(area=area, customer=user, project=project, start=start, end=end)
    if BillableItemType.TRAINING in types:
        TrainingSession.objects.create(
            trainee=user,
            trainer=staff,
            tool=tool,
            project=project,
            date=end,
            duration=30,
            type=TrainingSession.Type.INDIVIDUAL,
        )
    if BillableItemType.MISSED_RESERVATION in types:
        Reservation.objects.create(
            user=user,
            creator=user,
            tool=tool,
            project=project,
            missed=True,
            start=start,
            end=end,
            short_notice=False,
        )
    if BillableItemType.CUSTOM_CHARGE in types:
        CustomCharge.objects.create(
            name="custom",
            customer=user,
            creator=staff,
            project=project,
            date=end,
            amount=10,
            cap_eligible=cap_eligible_custom_charge,
        )



# Generated by Django 4.2.20 on 2025-04-28 14:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0008_alter_ratecategory_name"),
    ]

    def forward(apps, schema_editor):
        Customization = apps.get_model("NEMO", "Customization")
        Customization.objects.filter(name="allow_blank_rate").delete()

    def reverse(apps, schema_editor):
        pass

    operations = [
        migrations.AddField(
            model_name="rate",
            name="effective_date",
            field=models.DateField(blank=True, help_text="The date when this rate becomes effective.", null=True),
        ),
        migrations.RunPython(forward, reverse),
    ]



# Generated by Django 4.2.23 on 2025-10-03 16:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("quotes", "0002_quote_add_tax_and_number"),
    ]

    operations = [
        migrations.AddField(
            model_name="quoteitem",
            name="service_fee",
            field=models.DecimalField(
                blank=True, decimal_places=2, help_text="Service fee for this item", max_digits=8, null=True
            ),
        ),
    ]



from __future__ import annotations

import datetime
from _decimal import Decimal
from datetime import date
from logging import getLogger
from typing import Dict, List, Optional, Tuple

from NEMO.constants import CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.models import (
    BaseCategory,
    BaseModel,
    EmailNotificationType,
    Project,
    User,
)
from NEMO.typing import QuerySetType
from NEMO.utilities import EmailCategory, get_month_timeframe, send_mail
from dateutil.relativedelta import relativedelta
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator, validate_comma_separated_integer_list
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from NEMO_billing.customization import BillingCustomization
from NEMO_billing.invoices.customization import InvoiceCustomization
from NEMO_billing.invoices.models import BillableItemType, Invoice, InvoiceConfiguration, InvoiceSummaryItem
from NEMO_billing.invoices.processors import BillableItem
from NEMO_billing.invoices.utilities import display_amount
from NEMO_billing.models import CoreFacility
from NEMO_billing.prepayments.exceptions import (
    ProjectFundsExpiredException,
    ProjectFundsInactiveException,
    ProjectInsufficientFundsException,
)
from NEMO_billing.utilities import (
    Months,
    filter_date_year_month_gt,
    filter_date_year_month_lte,
    get_charges_amount_between,
    number_of_months_between_dates,
)


class ProjectPrepaymentDetail(BaseModel):
    project = models.OneToOneField(Project, verbose_name="Project", help_text="The project", on_delete=models.CASCADE)
    configuration = models.ForeignKey(InvoiceConfiguration, null=True, blank=True, on_delete=models.CASCADE)
    charge_types = models.CharField(
        validators=[validate_comma_separated_integer_list],
        max_length=CHAR_FIELD_SMALL_LENGTH,
        help_text="List of charge types allowed",
    )
    only_core_facilities = models.ManyToManyField(
        CoreFacility,
        blank=True,
        help_text="Limit which core facilities are allowed for this project. Leave blank to allow them all",
    )
    overdraft_amount_allowed = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="When set, this is the overdraft amount allowed for the project. Only works when there is at least one active fund with a positive balance.",
    )
    balance_last_updated = models.DateField(blank=True)

    @property
    def billable_charge_types(self) -> List[BillableItemType]:
        return [BillableItemType(int(value)) for value in self.charge_types.split(",") if value]

    def get_charge_types_display(self):
        return mark_safe(
            "<br>".join([charge_type.friendly_display_name() for charge_type in self.billable_charge_types])
        )

    def get_only_core_facilities_display(self):
        if not self.only_core_facilities.exists():
            return "All"
        return mark_safe("<br>".join([facility.name for facility in self.only_core_facilities.all()]))

    def active_funds(self, check_date: date, include_zero_balance=False) -> QuerySetType[Fund]:
        funds = self.fund_set.all()
        # Funds cannot all have expired
        non_expired_funds = funds.filter(
            Q(expiration_month__isnull=True, expiration_year__isnull=True)
            | filter_date_year_month_gt("expiration", check_date)
        )
        if not non_expired_funds.exists():
            raise ProjectFundsExpiredException(self.project)
        # At least one fund has to be active
        active_funds = non_expired_funds.filter(filter_date_year_month_lte("start", check_date))
        if not active_funds.exists():
            raise ProjectFundsInactiveException(self.project, check_date)
        if not include_zero_balance:
            active_funds = active_funds.filter(balance__gt=0)
        return active_funds.order_by("start_year", "start_month", "id")

    def update_balances_with_new_charges(self, new_charges: Decimal, as_of_date: date) -> List[Tuple[Fund, Decimal]]:
        fund_and_used_amount = []
        # This is used at the end of the month when invoicing
        active_funds_sorted_by_date = self.active_funds(as_of_date)
        last_item = active_funds_sorted_by_date.last()
        # Max out active funds one by one until the last one or no more charges
        for fund in active_funds_sorted_by_date:
            if new_charges <= 0:
                break
            if fund == last_item:
                # last fund, use all remaining charges
                update_amount = new_charges
            else:
                # max out other funds
                update_amount = min(fund.balance, new_charges)
            fund.update_balance(update_amount)
            fund_and_used_amount.append((fund, update_amount))
            new_charges = new_charges - update_amount
        self.balance_last_updated = as_of_date
        self.save()
        return fund_and_used_amount

    def get_prepayment_info(
        self, until: date, start_in_month: date = None, raise_exception=False
    ) -> (List[BillableItem], Decimal, Dict[int, Decimal]):
        # Returns total charges, total charges amount, and fund balances
        total_charges: List[BillableItem] = []
        total_charges_amount = Decimal(0)
        if not start_in_month:
            # balance is always one month behind
            start_in_month = (self.balance_last_updated + relativedelta(months=1)).replace(day=1)
        months = number_of_months_between_dates(until, start_in_month)
        # keep track of fund balances month to month
        fund_balances: Dict[int, Decimal] = {}
        for month in range(0, months + 1):
            month_date = start_in_month + relativedelta(months=month)
            # beginning and end of the month
            start, end = get_month_timeframe(month_date.isoformat())
            monthly_charges, monthly_charges_amount = get_charges_amount_between(
                self.project, self.configuration, start, end
            )
            total_charges.extend(monthly_charges)
            total_charges_amount = total_charges_amount + monthly_charges_amount
            if monthly_charges_amount:
                # only need to check funds valid at this date (expired or inactive won't be returned by active_funds)
                funds_left = Decimal(0)
                new_funds = Fund.objects.none()
                try:
                    new_funds = self.active_funds(end.date())
                except Exception:
                    if raise_exception:
                        raise
                last_fund_checked: Optional[Fund] = None
                last_item = new_funds.last()
                for fund in new_funds:
                    balance = max(fund_balances.setdefault(fund.id, fund.balance), Decimal(0))
                    if fund == last_item:
                        # last fund, use all remaining charges
                        update_amount = monthly_charges_amount
                    else:
                        # max out other funds
                        update_amount = min(fund.balance, monthly_charges_amount)
                    fund_balances[fund.id] = balance - update_amount
                    monthly_charges_amount = monthly_charges_amount - update_amount
                    funds_left += fund_balances[fund.id]
                    last_fund_checked = fund
                if last_fund_checked:
                    last_fund_checked.check_for_low_balance(funds_left)
                if (
                    not self.overdraft_amount_allowed
                    and not funds_left > 0
                    or self.overdraft_amount_allowed
                    and not funds_left + self.overdraft_amount_allowed > 0
                ):
                    if raise_exception:
                        raise ProjectInsufficientFundsException(self.project)
        total_charges.sort(key=lambda x: x.start)
        return total_charges, total_charges_amount, fund_balances

    def invoice_fund_summaries(self, invoice: Invoice) -> list[InvoiceSummaryItem]:
        fund_summaries = []
        fund_and_amounts = self.update_balances_with_new_charges(invoice.total_amount, invoice.end)
        for fund_and_amount in fund_and_amounts:
            fund, amount = fund_and_amount
            fund_summary = InvoiceSummaryItem(
                invoice=invoice, name=f"{fund} balance: {display_amount(fund.balance, invoice.configuration)}"
            )
            fund_summary.amount = -amount
            fund_summary.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.FUND
            # Set fund id in details, so we can find it later for voiding/deleting
            fund_summary.details = str(fund.id)
            fund_summaries.append(fund_summary)
        if InvoiceCustomization.get_bool("invoice_funds_show_total_balance"):
            # find all active funds, show total balance as type "other"
            # (otherwise it will break when trying to find fund by id)
            total_balance = Decimal(sum(fund.balance for fund in self.active_funds(invoice.end)))
            if total_balance > 0:
                total_balance_summary = InvoiceSummaryItem(
                    invoice=invoice,
                    name=f"Total funds remaining: {display_amount(total_balance, invoice.configuration)}",
                )
                total_balance_summary.amount = 0
                total_balance_summary.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.OTHER
                fund_summaries.append(total_balance_summary)
        return fund_summaries

    def restore_funds(self, invoice: Invoice, request):
        previous_balance_date = invoice.end - relativedelta(months=1)
        for fund_summary in invoice.invoicesummaryitem_set.filter(
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.FUND
        ):
            # Restore balance on each fund
            fund = Fund.objects.get(id=fund_summary.details)
            fund.update_balance(fund_summary.amount)
            messages.success(
                request,
                f"the balance for {fund} was successfully credited back {display_amount(-fund_summary.amount, invoice.configuration)}.",
                "data-speed=30000",
            )
        self.balance_last_updated = previous_balance_date
        self.save()

    def clean(self):
        if self.project_id:
            if not self.project.projectbillingdetails.no_tax and not self.configuration:
                raise ValidationError(
                    {
                        "configuration": _(
                            "Configuration is required for taxed projects. Select a configuration or make the project tax exempt"
                        )
                    }
                )

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if not self.pk and self.balance_last_updated is None:
            # Set original balance date to be last day of last month
            # Reusing get_month_timeframe since that's what will be used when running invoices
            self.balance_last_updated = get_month_timeframe(
                (datetime.date.today() - relativedelta(months=1)).isoformat()
            )[1]
        super().save(force_insert, force_update, using, update_fields)

    def __str__(self):
        return f"Prepayment details for {self.project.name}"

    class Meta:
        ordering = ["project"]


class FundType(BaseCategory):
    pass


class Fund(BaseModel):
    project_prepayment = models.ForeignKey(ProjectPrepaymentDetail, on_delete=models.CASCADE)
    fund_type = models.ForeignKey(FundType, on_delete=models.PROTECT)
    reference = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, null=True, blank=True)
    amount = models.DecimalField(decimal_places=2, max_digits=14)
    start_month = models.PositiveIntegerField(choices=Months.choices)
    start_year = models.PositiveIntegerField(
        validators=[MinValueValidator(1900), MaxValueValidator(datetime.MAXYEAR)],
    )
    expiration_month = models.PositiveIntegerField(null=True, blank=True, choices=Months.choices)
    expiration_year = models.PositiveIntegerField(
        null=True,
        blank=True,
        validators=[MinValueValidator(1900), MaxValueValidator(datetime.MAXYEAR)],
    )
    balance = models.DecimalField(blank=True, decimal_places=2, max_digits=14)
    balance_warning_percent = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="Send a warning email when the balance is below this percent.",
        validators=[MaxValueValidator(100)],
    )
    balance_warning_sent = models.DateTimeField(null=True, blank=True)
    note = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, null=True, blank=True)

    @property
    def start_date(self) -> date:
        if self.start_year and self.start_month:
            return date(self.start_year, self.start_month, 1)

    @start_date.setter
    def start_date(self, value: date):
        if value:
            self.start_month, self.start_year = value.month, value.year
        else:
            self.start_month, self.start_year = None, None

    @property
    def expiration_date(self) -> date:
        if self.expiration_year and self.expiration_month:
            return date(self.expiration_year, self.expiration_month, 1)

    @expiration_date.setter
    def expiration_date(self, value: date):
        if value:
            self.expiration_month, self.expiration_year = value.month, value.year
        else:
            self.expiration_month, self.expiration_year = None, None

    def is_active(self, check_date: date):
        return self in self.project_prepayment.active_funds(check_date)

    def update_balance(self, new_charges: Decimal):
        self.balance = self.balance - new_charges
        self.save()

    def check_for_low_balance(self, balance_left: Decimal, raise_exception=False):
        try:
            if not self.balance_warning_sent:
                if self.balance_warning_percent:
                    warning_amount = self.balance_warning_percent / Decimal(100) * self.amount
                    if balance_left <= warning_amount:
                        subject = f"Low fund balance for project {self.project_prepayment.project.name}"
                        # Send to accounting staff, billing accounting email and project email
                        recipients = [
                            email
                            for user in User.objects.filter(is_active=True, is_accounting_officer=True)
                            for email in user.get_emails(EmailNotificationType.BOTH_EMAILS)
                        ]
                        billing_email = BillingCustomization.get("billing_accounting_email_address")
                        if billing_email:
                            recipients.append(billing_email)
                        recipients.extend(self.project_prepayment.project.projectbillingdetails.email_to())
                        message = "Hello,<br><br>\n\n"
                        message += f"You project {self.project_prepayment.project.name} has a low fund balance:<br>\n"
                        message += f"Original amount: {self.amount:.2f}<br>\n"
                        message += f"Current balance: {balance_left:.2f}<br>\n"
                        send_mail(
                            subject=subject,
                            content=message,
                            from_email=billing_email,
                            to=recipients,
                            email_category=EmailCategory.GENERAL,
                        )
                        self.balance_warning_sent = timezone.now()
                        self.save(update_fields=["balance_warning_sent"])
        except:
            getLogger(__name__).exception("Error checking/sending low balance email")
            if raise_exception:
                raise

    def clean(self):
        errors = {}
        if self.balance and self.balance > self.amount:
            errors["balance"] = _("The balance cannot be greater than the fund amount")
        if self.start_month and not self.start_year or self.start_year and not self.start_month:
            errors["start_year"] = "Both year/month are required together"
            errors["start_month"] = "Both year/month are required together"
        if self.expiration_month and not self.expiration_year or self.expiration_year and not self.expiration_month:
            errors["expiration_year"] = "Both year/month are required together"
            errors["expiration_month"] = "Both year/month are required together"
        if errors:
            raise ValidationError(errors)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if not self.pk and self.balance is None:
            self.balance = self.amount
        super().save(force_insert, force_update, using, update_fields)

    def __str__(self):
        fund_name = f"Fund ref: {self.reference}" if self.reference else f"Fund #{self.id}"
        return f"{fund_name} ({self.fund_type})"



from NEMO.serializers import ModelSerializer
from NEMO.views.api import (
    ModelViewSet,
    boolean_filters,
    date_filters,
    datetime_filters,
    key_filters,
    manykey_filters,
    number_filters,
    string_filters,
)
from rest_flex_fields import FlexFieldsModelSerializer

from NEMO_billing.quotes.models import Quote, QuoteConfiguration, QuoteItem


class QuoteConfigurationSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = QuoteConfiguration
        fields = "__all__"


class QuoteSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = Quote
        fields = "__all__"
        expandable_fields = {
            "project": "NEMO_billing.invoices.api.ProjectWithDetailsSerializer",
            "configuration": "NEMO_billing.quotes.api.QuoteConfigurationSerializer",
            "users": ("NEMO.serializers.UserSerializer", {"many": True}),
            "creator": "NEMO.serializers.UserSerializer",
            "approved_by": "NEMO.serializers.UserSerializer",
        }


class QuoteItemSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = QuoteItem
        fields = "__all__"
        expandable_fields = {"quote": "NEMO_billing.quotes.api.QuoteSerializer"}


class QuoteConfigurationViewSet(ModelViewSet):
    filename = "quote_configurations"
    queryset = QuoteConfiguration.objects.all()
    serializer_class = QuoteConfigurationSerializer
    filterset_fields = {
        "id": key_filters,
        "name": string_filters,
        "expiration_in_days": number_filters,
        "current_quote_number": number_filters,
        "email_cc": string_filters,
        "merchant_name": string_filters,
        "merchant_details": string_filters,
        "terms": string_filters,
        "currency": string_filters,
        "currency_symbol": string_filters,
        "tax": number_filters,
        "tax_name": string_filters,
        "create_permissions": string_filters,
        "approval_permissions": string_filters,
    }


class QuoteViewSet(ModelViewSet):
    filename = "quotes"
    queryset = Quote.objects.all()
    serializer_class = QuoteSerializer
    filterset_fields = {
        "id": key_filters,
        "name": string_filters,
        "quote_number": string_filters,
        "project": key_filters,
        "configuration": key_filters,
        "users": manykey_filters,
        "emails": string_filters,
        "expiration_date": date_filters,
        "creator": key_filters,
        "approved_by": key_filters,
        "created_date": datetime_filters,
        "updated_date": datetime_filters,
        "published_date": datetime_filters,
        "last_emails_sent_date": datetime_filters,
        "status": string_filters,
        "add_tax": boolean_filters,
    }


class QuoteItemViewSet(ModelViewSet):
    filename = "quote_items"
    queryset = QuoteItem.objects.all()
    serializer_class = QuoteItemSerializer
    filterset_fields = {
        "id": key_filters,
        "quantity": number_filters,
        "amount": number_filters,
        "minimum_charge": number_filters,
        "service_fee": number_filters,
        "description": string_filters,
        "rate_type": number_filters,
        "quote": key_filters,
    }



from typing import Dict

from NEMO.decorators import customization
from NEMO.exceptions import InvalidCustomizationException
from NEMO.models import Tool
from NEMO.views.customization import CustomizationBase
from django.core.exceptions import ValidationError
from django.core.validators import validate_comma_separated_integer_list, validate_integer

from NEMO_billing.cap_discount.models import CAPDiscountConfiguration
from NEMO_billing.invoices.models import BillableItemType


@customization(key="cap_discount", title="Billing CAP Discounts")
class CAPDiscountCustomization(CustomizationBase):
    variables = {
        "cap_billing_exclude_tools": "",
        "cap_billing_default_billable_types": "",
        "cap_billing_default_interval": "1",
        "cap_billing_default_frequency": "",
    }

    def context(self) -> Dict:
        # Override to add list of tools
        dictionary = super().context()
        dictionary["tools"] = Tool.objects.all()
        dictionary["selected_tools"] = Tool.objects.filter(id__in=self.get_excluded_tool_ids())
        dictionary["billable_types"] = BillableItemType.choices()
        dictionary["selected_billable_types"] = self.get_default_billable_types()
        dictionary["recurrence_frequencies"] = CAPDiscountConfiguration.reset_frequency.field.choices
        return dictionary

    def validate(self, name, value):
        if name in ["cap_billing_exclude_tools", "cap_billing_default_billable_types"] and value:
            validate_comma_separated_integer_list(value)
        if name in ["cap_billing_default_frequency", "cap_billing_default_interval"] and value:
            validate_integer(value)

    def save(self, request, element=None) -> Dict[str, Dict[str, str]]:
        errors = super().save(request, element)
        exclude_tools = ",".join(request.POST.getlist("cap_billing_exclude_tools_list", []))
        try:
            self.validate("cap_billing_exclude_tools", exclude_tools)
            type(self).set("cap_billing_exclude_tools", exclude_tools)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["cap_billing_exclude_tools"] = {"error": str(e.message or e.msg), "value": exclude_tools}
        default_types = ",".join(request.POST.getlist("cap_billing_default_billable_types_list", []))
        try:
            self.validate("cap_billing_default_billable_types", default_types)
            type(self).set("cap_billing_default_billable_types", default_types)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["cap_billing_default_billable_types"] = {"error": str(e.message or e.msg), "value": default_types}
        return errors

    @classmethod
    def get_excluded_tool_ids(cls):
        return [int(pk) for pk in cls.get("cap_billing_exclude_tools").split(",") if pk]

    @classmethod
    def get_default_billable_types(cls):
        return [int(value) for value in cls.get("cap_billing_default_billable_types").split(",") if value]



# Generated by Django 3.2.20 on 2023-09-07 13:41

import django.db.models.deletion
from django.db import migrations, models


class MysqlExemptRemoveConstraint(migrations.RemoveConstraint):
    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        if schema_editor.connection.vendor.startswith("mysql"):
            return
        super().database_forwards(app_label, schema_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        if schema_editor.connection.vendor.startswith("mysql"):
            return
        super().database_backwards(app_label, schema_editor, from_state, to_state)


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0005_version_2_5_6"),
        ("cap_discount", "0004_shortened_unique_constraints"),
    ]

    def migrate_configuration_core_facilities(apps, schema_editor):
        CAPDiscountConfiguration = apps.get_model("cap_discount", "CAPDiscountConfiguration")

        for config in CAPDiscountConfiguration.objects.all():
            config.core_facilities.add(config.core_facility)

    def migrate_configuration_core_facilities_back(apps, schema_editor):
        CAPDiscountConfiguration = apps.get_model("cap_discount", "CAPDiscountConfiguration")

        for config in CAPDiscountConfiguration.objects.all():
            config.core_facility = config.core_facilities.first()
            config.save(update_fields=["core_facility"])

    operations = [
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="core_facility",
            field=models.ForeignKey(
                blank=True,
                help_text="The core facility this CAP applies to",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cf",
                to="NEMO_billing.corefacility",
            ),
        ),
        # We need to drop the index and constraint temporarily so MySQL is happy removing the unique constraint
        # This solves the Cannot drop index '...': needed in a foreign key constraint
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="rate_category",
            field=models.ForeignKey(
                help_text="The rate category this CAP applies to",
                on_delete=django.db.models.deletion.CASCADE,
                to="rates.ratecategory",
                db_index=False,
                db_constraint=False,
            ),
        ),
        migrations.AddField(
            model_name="capdiscountconfiguration",
            name="core_facilities",
            field=models.ManyToManyField(
                blank=True,
                help_text="The core facility(ies) this CAP applies to. If multiple are selected, the CAP is shared between facilities",
                to="NEMO_billing.CoreFacility",
            ),
        ),
        migrations.RemoveConstraint(
            model_name="capdiscountconfiguration",
            name="capdiscountconfiguration_unique_rate_category_facility",
        ),
        MysqlExemptRemoveConstraint(
            model_name="capdiscountconfiguration",
            name="capdiscountconfiguration_unique_rate_category_null_facility",
        ),
        migrations.RunPython(migrate_configuration_core_facilities, migrate_configuration_core_facilities_back),
        migrations.RemoveField(
            model_name="capdiscountconfiguration",
            name="core_facility",
        ),
        # Put back db index and constraint now
        migrations.AlterField(
            model_name="capdiscountconfiguration",
            name="rate_category",
            field=models.ForeignKey(
                help_text="The rate category this CAP applies to",
                on_delete=django.db.models.deletion.CASCADE,
                to="rates.ratecategory",
            ),
        ),
    ]



import datetime

from NEMO.models import Consumable, Tool
from django.core.exceptions import ValidationError
from django.test import TestCase

from NEMO_billing.rates.models import DailySchedule, Rate, RateCategory, RateTime, RateType


class RateTestCase(TestCase):
    def test_rate_no_category_no_items(self):
        # Tool missed reservation is a good example of a rate not category or item specific
        tool_missed_resa = RateType.objects.get(type="TOOL_MISSED_RESERVATION")
        self.assertFalse(tool_missed_resa.category_specific)
        self.assertFalse(tool_missed_resa.item_specific)
        Rate.objects.create(type=tool_missed_resa, amount="10.00")
        # Try creating another one (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=tool_missed_resa, amount="20.00").full_clean()
            self.assertIn("__all__", cm.exception.error_dict)

        # Create another one with an effective date
        effective_date = datetime.date(2023, 1, 1)
        effective_date_2 = datetime.date(2024, 1, 1)
        Rate.objects.create(type=tool_missed_resa, amount="10.00", effective_date=effective_date)
        Rate.objects.create(type=tool_missed_resa, amount="10.00", effective_date=effective_date_2)

        # Try creating another one with the same effective date (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=tool_missed_resa, amount="20.00", effective_date=effective_date).full_clean()
            self.assertIn("__all__", cm.exception.error_dict)

    def test_rate_no_category_items(self):
        # Consumable is a good example of a rate not category specific but item specific
        gold = Consumable.objects.create(name="Gold", quantity=10, reminder_email="test@test.com", reminder_threshold=5)
        platinum = Consumable.objects.create(
            name="Platinum", quantity=10, reminder_email="test@test.com", reminder_threshold=5
        )
        consumable_type = RateType.objects.get(type="CONSUMABLE")
        self.assertFalse(consumable_type.category_specific)
        self.assertTrue(consumable_type.item_specific)
        Rate.objects.create(type=consumable_type, consumable=gold, amount="10.00")
        # Try creating another one (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=consumable_type, consumable=gold, flat=True, amount="20.00").full_clean()
        self.assertIn("__all__", cm.exception.error_dict)
        # Try creating another one with another item (should pass)
        Rate(type=consumable_type, consumable=platinum, flat=True, amount="20.00").full_clean()

        # Create another one with an effective date
        effective_date = datetime.date(2023, 1, 1)
        effective_date_2 = datetime.date(2024, 1, 1)
        Rate.objects.create(type=consumable_type, consumable=gold, amount="10.00", effective_date=effective_date)
        Rate.objects.create(type=consumable_type, consumable=gold, amount="10.00", effective_date=effective_date_2)

        # Try creating another one with the same effective date (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=consumable_type, consumable=gold, amount="20.00", effective_date=effective_date).full_clean()
            self.assertIn("__all__", cm.exception.error_dict)

    def test_rate_category_no_items(self):
        # Staff charge is a good example of a rate category specific but not item specific
        industry = RateCategory.objects.create(name="Industry")
        academia = RateCategory.objects.create(name="Academia")
        staff_charge_type = RateType.objects.get(type="STAFF_CHARGE")
        self.assertTrue(staff_charge_type.category_specific)
        self.assertFalse(staff_charge_type.item_specific)
        Rate.objects.create(type=staff_charge_type, category=industry, amount="10.00")
        # Try creating another one (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=staff_charge_type, category=industry, amount="20.00").full_clean()
        self.assertIn("__all__", cm.exception.error_dict)
        # Try creating another one with another category (should pass)
        Rate(type=staff_charge_type, category=academia, amount="20.00").full_clean()

        # Create another one with an effective date
        effective_date = datetime.date(2023, 1, 1)
        effective_date_2 = datetime.date(2024, 1, 1)
        Rate.objects.create(type=staff_charge_type, category=industry, amount="10.00", effective_date=effective_date)
        Rate.objects.create(type=staff_charge_type, category=industry, amount="10.00", effective_date=effective_date_2)

        # Try creating another one with the same effective date (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=staff_charge_type, category=industry, amount="20.00", effective_date=effective_date).full_clean()
            self.assertIn("__all__", cm.exception.error_dict)

    def test_rate_category_items(self):
        # Tool usage is a good example of a rate category specific and item specific
        tool1 = Tool.objects.create(name="tool_1")
        tool2 = Tool.objects.create(name="tool_2")
        industry = RateCategory.objects.create(name="Industry")
        academia = RateCategory.objects.create(name="Academia")
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        self.assertTrue(tool_usage_type.category_specific)
        self.assertTrue(tool_usage_type.item_specific)
        Rate.objects.create(type=tool_usage_type, tool=tool1, category=industry, amount="10.00")
        # Try creating another one (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=tool_usage_type, tool=tool1, category=industry, amount="20.00").full_clean()
        self.assertIn("__all__", cm.exception.error_dict)
        # Try creating another one with another category (should pass)
        Rate(type=tool_usage_type, tool=tool1, category=academia, amount="20.00").full_clean()
        # Try creating another one with another item (should pass)
        Rate(type=tool_usage_type, tool=tool2, category=industry, amount="30.00").full_clean()

        # Create another one with an effective date
        effective_date = datetime.date(2023, 1, 1)
        effective_date_2 = datetime.date(2024, 1, 1)
        Rate.objects.create(
            type=tool_usage_type, tool=tool1, category=industry, amount="10.00", effective_date=effective_date
        )
        Rate.objects.create(
            type=tool_usage_type, tool=tool1, category=industry, amount="10.00", effective_date=effective_date_2
        )

        # Try creating another one with the same effective date (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(
                type=tool_usage_type, tool=tool1, category=industry, amount="20.00", effective_date=effective_date
            ).full_clean()
            self.assertIn("__all__", cm.exception.error_dict)

    def test_rate_time_overlap(self):
        # Tool usage is a good example of a rate category specific and item specific
        tool1 = Tool.objects.create(name="tool_1")
        tool_usage_type = RateType.objects.get(type="TOOL_USAGE")
        self.assertTrue(tool_usage_type.item_specific)
        # Create base rate
        Rate.objects.create(type=tool_usage_type, tool=tool1, amount="10.00")

        # Create first rate time
        rate_time_1 = RateTime.objects.create(name="Mon 10AM - 10PM")
        DailySchedule.objects.create(
            rate_time=rate_time_1, start_time=datetime.time(10, 0, 0), end_time=datetime.time(22, 0, 0), weekday=0
        )
        # Create overlapping time
        rate_time_2 = RateTime.objects.create(name="Mon 8PM - 10AM")
        daily_schedule_2 = DailySchedule.objects.create(
            rate_time=rate_time_2, start_time=datetime.time(20, 0, 0), end_time=datetime.time(10, 0, 0), weekday=0
        )

        rate_1 = Rate(type=tool_usage_type, tool=tool1, amount="10.00", time=rate_time_1)
        self.assertIsNone(rate_1.full_clean())
        rate_1.save()

        # Try creating another one (should fail)
        with self.assertRaises(ValidationError) as cm:
            Rate(type=tool_usage_type, tool=tool1, amount="20.00", time=rate_time_2).full_clean()

        # Create another one that doesn't overlap (start at 11PM)
        daily_schedule_2.start_time = datetime.time(23, 0, 0)
        self.assertIsNone(daily_schedule_2.full_clean())
        daily_schedule_2.save()
        rate_2 = Rate(type=tool_usage_type, tool=tool1, amount="20.00", time=rate_time_2)
        self.assertIsNone(rate_2.full_clean())
        rate_2.save()
        self.assertIsNone(rate_2.full_clean())

        # Changing the rate time later should trigger validation error if it will make rates overlap
        daily_schedule_2.start_time = datetime.time(20, 0, 0)
        with self.assertRaises(ValidationError) as cm:
            daily_schedule_2.full_clean()



# Generated by Django 2.2.21 on 2021-06-07 02:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0003_version_1_4_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="projectbillingdetails",
            name="contact_phone",
            field=models.CharField(blank=True, help_text="The contact's phone number", max_length=40, null=True),
        ),
    ]



# Generated by Django 4.2.23 on 2025-11-07 20:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0014_invoicedetailitem_waived"),
    ]

    operations = [
        migrations.AddField(
            model_name="projectbillingdetails",
            name="no_cap",
            field=models.BooleanField(
                default=False, help_text="Check this box if this project should not count towards CAP."
            ),
        ),
    ]



# Generated by Django 2.2.20 on 2021-04-22 12:21

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO", "0027_version_3_8_0"),
        ("NEMO_billing", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="corerelationship",
            name="staff_charge",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="core_rel",
                to="NEMO.StaffCharge",
            ),
        ),
    ]



# Generated by Django 2.2.24 on 2021-09-02 17:42

from django.db import migrations, models

import NEMO_billing.invoices.utilities


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0005_version_1_6_1"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="file",
            field=models.FileField(
                blank=True,
                max_length=255,
                null=True,
                upload_to=NEMO_billing.invoices.utilities.get_invoice_document_filename,
            ),
        ),
    ]



# Generated by Django 3.2.25 on 2024-05-27 15:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0011_version_2_4_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoiceconfiguration",
            name="invoice_title",
            field=models.CharField(
                default="Invoice",
                help_text="The title displayed on the first page of the invoice (e.g. Invoice/Statement/Quote)",
                max_length=255,
            ),
        ),
    ]



from NEMO.admin import ProjectAdmin
from NEMO.models import Project, User
from NEMO.utilities import format_datetime, new_model_copy
from django.apps import apps
from django.contrib import admin
from django.contrib.admin import SimpleListFilter, register
from django.contrib.contenttypes.models import ContentType
from django.db.models.functions import Length
from django.urls import NoReverseMatch, reverse
from django.utils.html import format_html

from NEMO_billing.invoices.models import (
    Invoice,
    InvoiceConfiguration,
    InvoiceDetailItem,
    InvoicePayment,
    InvoiceSummaryItem,
    ProjectBillingDetails,
)
from NEMO_billing.invoices.processors import invoice_data_processor_class as data_processor
from NEMO_billing.invoices.views.invoices import (
    delete_invoice,
    mark_invoice_paid_in_full,
    review_invoice,
    send_invoice,
    void_invoice,
    zip_response,
)
from NEMO_billing.templatetags.billing_tags import cap_discount_installed
from NEMO_billing.utilities import hide_form_field


def clone_project(project: Project):
    """
    Create a clone of selected projects and save them.

    :param project:
    :return: Newly cloned project
    """
    project_details = getattr(project, "projectbillingdetails", None)
    # Only copy active users and PIs
    project_users = project.user_set.filter(is_active=True)
    project_pis = project.manager_set.filter(is_active=True)
    only_allow_tools = project.only_allow_tools.all()
    new_project = new_model_copy(project)
    new_project.name = f"Copy of {project.name}"
    new_project.save()
    # Set relations
    new_project.user_set.set(project_users)
    new_project.manager_set.set(project_pis)
    new_project.only_allow_tools.set(only_allow_tools)
    if project_details:
        new_project_details = new_model_copy(project_details)
        new_project_details.project = new_project
        new_project_details.save()
    return project


@admin.action(description="Clone selected projects")
def clone_projects(modeladmin, request, queryset):
    for project in queryset.all():
        clone_project(project)


def clone_configuration(configuration: InvoiceConfiguration):
    """
    Create a clone of selected configurations and save them.

    :param configuration:
    :return: Newly cloned configuration
    """
    new_configuration = new_model_copy(configuration)
    new_configuration.name = f"Copy of {configuration.name}"
    new_configuration.save()

    return new_configuration


@admin.action(description="Clone selected Configurations")
def clone_configurations(modeladmin, request, queryset):
    for configuration in queryset.all():
        clone_configuration(configuration)


@admin.action(description="Email selected invoices to customers")
def email_invoices(modeladmin, request, queryset):
    for invoice in queryset.all():
        send_invoice(request, invoice.id)


@admin.action(description="Mark selected invoices as reviewed")
def review_invoices(modeladmin, request, queryset):
    for invoice in queryset.all():
        review_invoice(request, invoice.id)


@admin.action(description="Mark selected invoices as reviewed and send them")
def review_and_send_invoices(modeladmin, request, queryset):
    for invoice in queryset.all():
        review_invoice(request, invoice.id)
        send_invoice(request, invoice.id)


@admin.action(description="Mark selected invoices as void")
def void_invoices(modeladmin, request, queryset):
    for invoice in queryset.all():
        void_invoice(request, invoice.id)


@admin.action(description="Download selected invoices")
def zip_invoices(modeladmin, request, queryset):
    return zip_response(request, queryset.all())


@admin.action(description="Export invoice data in CSV")
def zip_invoices_csv(modeladmin, request, queryset):
    return zip_response(request, queryset.all(), "csv")


@admin.action(description="Mark selected invoices as paid in full")
def mark_invoices_as_paid(modeladmin, request, queryset):
    for invoice in queryset.all():
        mark_invoice_paid_in_full(request, invoice.id)


class InvoiceSummaryItemInline(admin.TabularInline):
    model = InvoiceSummaryItem
    can_delete = False

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


class InvoiceDetailItemInline(admin.TabularInline):
    model = InvoiceDetailItem
    can_delete = False
    exclude = ("content_type", "object_id")
    readonly_fields = ("get_item",)
    fields = ("get_item", "core_facility", "item_type", "name", "quantity", "start", "end", "user", "rate", "amount")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False

    @admin.display(description="Item")
    def get_item(self, obj: InvoiceDetailItem):
        if not obj.content_type or not obj.object_id:
            return "-"
        app_label, model = obj.content_type.app_label, obj.content_type.model
        viewname = f"admin:{app_label}_{model}_change"
        try:
            args = [obj.object_id]
            link = reverse(viewname, args=args)
        except NoReverseMatch:
            return "-"
        else:
            return format_html('<a href="{}">{} - #{}</a>', link, self.get_model_name(obj.content_type), obj.object_id)

    def get_model_name(self, content_type: ContentType):
        try:
            model = apps.get_model(content_type.app_label, content_type.model)
            return model._meta.verbose_name.capitalize()
        except (LookupError, AttributeError):
            return ""


class InvoiceStatusFilter(SimpleListFilter):
    title = "status"
    parameter_name = "status"

    def lookups(self, request, model_admin):
        return [
            ("unsent-reviewed", "Reviewed, not sent"),
            ("sent", "Sent"),
            ("unsent", "Not sent"),
            ("reviewed", "Reviewed"),
            ("unreviewed", "Not reviewed"),
            ("voided", "Voided"),
        ]

    def queryset(self, request, queryset):
        new_queryset = Invoice.objects.filter(voided_date=None)
        if self.value() in ["unsent-reviewed", "unsent"]:
            new_queryset = new_queryset.filter(sent_date=None)
        if self.value() in ["reviewed", "unsent-reviewed"]:
            new_queryset = new_queryset.filter(reviewed_date__isnull=False)
        if self.value() == "sent":
            new_queryset = new_queryset.filter(sent_date__isnull=False)
        if self.value() == "unreviewed":
            new_queryset = new_queryset.filter(reviewed_date=None)
        if self.value() == "voided":
            new_queryset = Invoice.objects.filter(voided_date__isnull=False)
        return new_queryset


@register(Invoice)
class InvoiceAdmin(admin.ModelAdmin):
    inlines = (InvoiceSummaryItemInline, InvoiceDetailItemInline)
    list_display = (
        "invoice_number",
        "get_invoice_date",
        "get_total_amount_currency",
        "get_tax_display",
        "project_details",
        "configuration",
        "created_date",
        "due_date",
        "last_sent_date",
        "reviewed_date",
        "voided_date",
        "file",
    )
    list_filter = (
        InvoiceStatusFilter,
        "start",
        "due_date",
        "configuration__currency",
        "project_details__project__name",
    )
    ordering = (Length("invoice_number").desc(), "-invoice_number")
    readonly_fields = ("voided_date", "voided_by", "created_date", "created_by")
    date_hierarchy = "start"
    search_fields = ["invoice_number", "project_details__project__name"]
    actions = (email_invoices, review_invoices, void_invoices, mark_invoices_as_paid, zip_invoices, zip_invoices_csv)

    @admin.display(description="Date", ordering="start")
    def get_invoice_date(self, obj: Invoice):
        return format_datetime(obj.start, "F Y")

    @admin.display(description="Amount", ordering="total_amount")
    def get_total_amount_currency(self, obj: Invoice):
        return obj.total_amount_display()

    @admin.display(description="Tax")
    def get_tax_display(self, obj: Invoice):
        return obj.tax_display()

    def delete_queryset(self, request, queryset):
        for invoice in queryset:
            if Invoice.objects.filter(id=invoice.id).exists():
                delete_invoice(request, invoice.id)

    def delete_model(self, request, obj: Invoice):
        delete_invoice(request, obj.id)

    def get_deleted_objects(self, objs, request):
        return data_processor.get_deleted_objects(objs, request, self.admin_site)


@register(InvoiceConfiguration)
class InvoiceConfigurationAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "merchant_name", "currency")
    ordering = ("name",)
    actions = (clone_configurations,)
    fieldsets = (
        (None, {"fields": ("name", "invoice_due_in", "invoice_title", "detailed_invoice", "hide_zero_charge")}),
        ("Communication", {"fields": ("email_from", "email_cc", "reminder_frequency")}),
        ("Merchant info", {"fields": ("merchant_name", "merchant_details", "merchant_logo", "terms")}),
        ("Currency", {"fields": ("currency", "currency_symbol")}),
        ("Tax", {"fields": ("tax_name", "tax")}),
        (
            "Display charges",
            {
                "fields": (
                    "separate_tool_usage_charges",
                    "separate_area_access_charges",
                    "separate_staff_charges",
                    "separate_consumable_charges",
                    "separate_missed_reservation_charges",
                    "separate_training_charges",
                    "separate_custom_charges",
                )
            },
        ),
    )


@register(InvoicePayment)
class InvoicePaymentAdmin(admin.ModelAdmin):
    readonly_fields = ("created_by", "created_date", "updated_by", "updated_date")
    list_display = (
        "invoice",
        "payment_received",
        "payment_processed",
        "get_amount_display",
        "updated_by",
        "updated_date",
    )
    list_filter = ("payment_received", "payment_processed", "invoice__project_details__project__name", "invoice")
    date_hierarchy = "payment_received"
    autocomplete_fields = ["invoice"]

    @admin.display(description="Amount", ordering="amount")
    def get_amount_display(self, obj: InvoicePayment):
        return obj.amount_display()

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created_by = request.user
        obj.updated_by = request.user
        obj.save()


class ProjectBillingDetailsInline(admin.StackedInline):
    model = ProjectBillingDetails
    can_delete = False
    min_num = 1
    verbose_name_plural = "details"

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj, **kwargs)
        if not cap_discount_installed():
            hide_form_field(formset.form, "no_cap", "base_fields")
        return formset

    def get_field_queryset(self, db, db_field, request):
        query_set = super().get_field_queryset(db, db_field, request)
        if db_field.name == "staff_host":
            query_set = query_set or User.objects.all()
            query_set = query_set.filter(is_active=True, is_staff=True)
        return query_set


class NewProjectAdmin(ProjectAdmin):
    actions = list(ProjectAdmin.actions) + [clone_projects]
    inlines = list(ProjectAdmin.inlines) + [ProjectBillingDetailsInline]


# Re-register ProjectAdmin
admin.site.unregister(Project)
admin.site.register(Project, NewProjectAdmin)



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

# Set some app default settings
for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NEMOInvoicesConfig(AppConfig):
    name = "NEMO_billing.invoices"
    verbose_name = "Billing Invoices"
    default_auto_field = "django.db.models.AutoField"



from __future__ import annotations

import os
from datetime import timedelta
from decimal import Decimal
from enum import Enum, unique
from logging import getLogger
from typing import Dict, List

from NEMO import fields
from NEMO.constants import CHAR_FIELD_LARGE_LENGTH, CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.models import BaseCategory, BaseModel, Customization, Project, SerializationByNameModel, User
from NEMO.utilities import create_email_attachment, quiet_int
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db import models
from django.db.models import Case, Q, Sum, When
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify

from NEMO_billing.invoices.utilities import (
    category_name_for_item_type,
    display_amount,
    export_invoice_filename,
    get_invoice_document_filename,
    get_merchant_logo_filename,
    render_and_send_email,
)
from NEMO_billing.rates.models import Rate, RateCategory, RateType


@unique
class BillableItemType(Enum):
    TOOL_USAGE = 1
    AREA_ACCESS = 2
    CONSUMABLE = 3
    MISSED_RESERVATION = 4
    STAFF_CHARGE = 5
    TRAINING = 6
    CUSTOM_CHARGE = 7

    @classmethod
    def choices(cls):
        return (
            (cls.TOOL_USAGE.value, "tool_usage"),
            (cls.AREA_ACCESS.value, "area_access"),
            (cls.CONSUMABLE.value, "consumable"),
            (cls.MISSED_RESERVATION.value, "missed_reservation"),
            (cls.STAFF_CHARGE.value, "staff_charge"),
            (cls.TRAINING.value, "training_session"),
            (cls.CUSTOM_CHARGE.value, "custom_charge"),
        )

    @classmethod
    def choices_except(cls, *billable_types):
        return [item for item in cls.choices() if BillableItemType(item[0]) not in billable_types]

    def is_time_type(self):
        return self in [
            BillableItemType.TOOL_USAGE,
            BillableItemType.AREA_ACCESS,
            BillableItemType.TRAINING,
            BillableItemType.STAFF_CHARGE,
            BillableItemType.MISSED_RESERVATION,
        ]

    def display_name(self):
        choices_as_dict = dict(self.choices())
        return choices_as_dict.get(self.value)

    def friendly_display_name(self):
        return self.display_name().replace("_", " ").capitalize()

    def category_name_for_item_type(self):
        return category_name_for_item_type(self)


class ProjectBillingDetails(BaseModel):
    project = models.OneToOneField(Project, on_delete=models.CASCADE)
    category = models.ForeignKey(RateCategory, null=True, blank=True, on_delete=models.SET_NULL)
    project_name = models.CharField(
        null=True,
        blank=True,
        max_length=CHAR_FIELD_LARGE_LENGTH,
        help_text="The project name that will appear on the invoices. Leave blank to use NEMO project name",
    )
    contact_name = models.CharField(
        null=True,
        blank=True,
        max_length=CHAR_FIELD_MEDIUM_LENGTH,
        help_text="The contact name to use in the invoice email",
    )
    contact_phone = models.CharField(null=True, blank=True, max_length=40, help_text="The contact's phone number")
    contact_email = fields.MultiEmailField(
        null=True,
        blank=True,
        help_text="Email to send the invoice to. A comma-separated list can be used. Leave blank to use project managers/PIs emails",
    )
    expires_on = models.DateField(
        null=True, blank=True, help_text="Date after which this project will be automatically deactivated."
    )
    addressee = models.TextField(null=True, blank=True, help_text="The addressee details to be included in the invoice")
    institution = models.ForeignKey(
        "NEMO_billing.Institution",
        null=True,
        blank=True,
        help_text="The project institution",
        on_delete=models.SET_NULL,
    )
    department = models.ForeignKey(
        "NEMO_billing.Department", null=True, blank=True, help_text="The project department", on_delete=models.SET_NULL
    )
    staff_host = models.ForeignKey(User, null=True, blank=True, help_text="The project host", on_delete=models.SET_NULL)
    comments = models.TextField(null=True, blank=True)
    no_charge = models.BooleanField(
        default=False, help_text="Check this box if invoices should not be created for this project."
    )
    no_tax = models.BooleanField(default=False, help_text="Check this box if this project is tax exempt.")
    no_cap = models.BooleanField(
        default=False, help_text="Check this box if this project should not count towards CAP."
    )

    @property
    def name(self):
        return self.project_name if self.project_name else self.project.name

    def clean(self):
        if not self.category and RateCategory.objects.exists():
            raise ValidationError({"category": "You need to select a rate category for this project"})

    def email_to(self) -> List[str]:
        # return project PIs emails if not set here
        return self.contact_email if not self.email_empty() else [pi.email for pi in self.project.manager_set.all()]

    def email_empty(self):
        return not self.contact_email or not [email for email in self.contact_email if email]

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = "Project details"


class InvoiceConfiguration(BaseModel):
    name = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True, help_text="The name of this invoice configuration"
    )
    invoice_due_in = models.PositiveIntegerField(
        help_text="The default number of days invoices are due after", default=30
    )
    invoice_title = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH,
        default="Invoice",
        help_text="The title displayed on the first page of the invoice (e.g. Invoice/Statement/Quote)",
    )
    reminder_frequency = models.PositiveIntegerField(
        null=True,
        blank=True,
        default=7,
        help_text="How often to send a reminder. Default value is 7, meaning every week after past due invoice",
    )
    email_from = models.EmailField(help_text="The email address used to send invoices and reminders")
    email_cc = fields.MultiEmailField(
        null=True, blank=True, help_text="Email to cc the invoice to. A comma-separated list can be used"
    )
    merchant_name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)
    merchant_details = models.TextField(
        null=True,
        blank=True,
        help_text="The merchant details to be included in the invoice (address, phone number etc.)",
    )
    merchant_logo = models.ImageField(null=True, blank=True, upload_to=get_merchant_logo_filename)
    terms = models.TextField(null=True, blank=True, help_text="Terms and conditions to be included in the invoice")

    currency = models.CharField(max_length=4, default="USD")
    currency_symbol = models.CharField(null=True, blank=True, max_length=4, default="$")

    tax = models.DecimalField(
        null=True, blank=True, decimal_places=3, max_digits=5, help_text="Tax in percent. For 20.5% enter 20.5"
    )
    tax_name = models.CharField(max_length=CHAR_FIELD_SMALL_LENGTH, null=True, blank=True, default="VAT")

    detailed_invoice = models.BooleanField(
        default=True, help_text="Check this box if customers should receive a detailed invoice."
    )
    hide_zero_charge = models.BooleanField(default=True, help_text="Hide charges with an amount equal to 0.")

    separate_tool_usage_charges = models.BooleanField(
        default=False, help_text="Check this box to display tool usage charges as a separate invoice category"
    )
    separate_area_access_charges = models.BooleanField(
        default=False, help_text="Check this box to display area access charges as a separate invoice category"
    )
    separate_staff_charges = models.BooleanField(
        default=False, help_text="Check this box to display staff charges as a separate invoice category"
    )
    separate_consumable_charges = models.BooleanField(
        default=False, help_text="Check this box to display supplies/material as a separate invoice category"
    )
    separate_missed_reservation_charges = models.BooleanField(
        default=False, help_text="Check this box to display missed reservation charges as a separate invoice category"
    )
    separate_training_charges = models.BooleanField(
        default=False, help_text="Check this box to display training charges as a separate invoice category"
    )
    separate_custom_charges = models.BooleanField(
        default=False, help_text="Check this box to display custom charges as a separate invoice category"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rate_help_initialized = False

    def tax_display(self):
        return f"{self.tax:.2f}"

    def tax_amount(self):
        return (self.tax or Decimal(0)) / Decimal(100)

    @property
    def rates_exist(self):
        if not self.rate_help_initialized:
            self.init_rate_help()
        return self._rates_exist

    @property
    def rates_types(self):
        if not self.rate_help_initialized:
            self.init_rate_help()
        return self._rates_types

    @property
    def all_rates(self):
        if not self.rate_help_initialized:
            self.init_rate_help()
        return self._all_rates

    def init_rate_help(self, extra_filter=Q()):
        # Call this to initiate a few helping rate items to basically cache them
        self._rates_types = {rate_type.type: rate_type for rate_type in RateType.objects.all()}
        self._rates_exist = RateCategory.objects.exists()
        self._all_rates = list(
            Rate.non_deleted()
            .filter(extra_filter)
            .select_related("type", "time", "category")
            .prefetch_related("time__dailyschedule_set")
        )
        self.rate_help_initialized = True

    def __str__(self):
        return self.name

    @classmethod
    def first_or_default(cls):
        try:
            if cls.objects.exists():
                return cls.objects.first()
        except Exception as e:
            getLogger(__name__).warning(e)
        return cls(currency="", currency_symbol="")


class Invoice(BaseModel):
    invoice_number = models.CharField(
        null=False,
        blank=True,
        max_length=CHAR_FIELD_SMALL_LENGTH,
        unique=True,
        help_text="Leave blank to be assigned automatically",
    )
    start = models.DateTimeField()
    end = models.DateTimeField()

    configuration = models.ForeignKey(InvoiceConfiguration, on_delete=models.PROTECT)
    project_details = models.ForeignKey(ProjectBillingDetails, on_delete=models.PROTECT)

    due_date = models.DateField(blank=True, null=True)
    sent_date = models.DateTimeField(blank=True, null=True)
    last_sent_date = models.DateTimeField(blank=True, null=True)
    last_reminder_sent_date = models.DateTimeField(blank=True, null=True)
    reviewed_date = models.DateTimeField(blank=True, null=True)
    reviewed_by = models.ForeignKey(
        User, blank=True, null=True, related_name="reviewed_invoice_set", on_delete=models.PROTECT
    )
    voided_date = models.DateTimeField(blank=True, null=True)
    voided_by = models.ForeignKey(
        User, blank=True, null=True, related_name="voided_invoice_set", on_delete=models.PROTECT
    )

    created_by = models.ForeignKey(User, on_delete=models.PROTECT)
    created_date = models.DateTimeField(auto_now_add=True)

    total_amount = models.DecimalField(decimal_places=2, max_digits=14)

    file = models.FileField(
        null=True, blank=True, max_length=CHAR_FIELD_MEDIUM_LENGTH, upload_to=get_invoice_document_filename
    )

    def generate_invoice_number(self, update: bool = False):
        from NEMO_billing.invoices.customization import InvoiceCustomization

        number_format = InvoiceCustomization.get("invoice_number_format")
        try:
            # replace with get use_cache=False when NEMO 7.3.x is released
            current_number = quiet_int(Customization.objects.get(name="invoice_number_current").value, 0)
        except Customization.DoesNotExist:
            current_number = 0
        current_number += 1
        if update:
            InvoiceCustomization.set("invoice_number_current", str(current_number))
        return number_format.format(current_number)

    def save(self, *args, **kwargs):
        if not self.invoice_number:
            self.invoice_number = self.generate_invoice_number(True)
        super().save(*args, **kwargs)

    def save_all(self, detail_items, summary_items):
        # Save invoice, detail items and summary items
        self.save()
        for detail_item in detail_items:
            detail_item.invoice = self
            detail_item.save(force_insert=True)
        for summary_item in summary_items:
            summary_item.invoice = self
            summary_item.save(force_insert=True)
        # We want to generate/render the invoice now and save the file so that it is set once and for all
        # (rather than generating it on the fly which would potentially use newer project info and config)
        # if an error rendering happens, the file will need to be regenerated
        self.render_and_save_file()

    def get_absolute_url(self):
        return reverse("view_invoice", kwargs={"invoice_id": self.id})

    def filename(self, extension=None):
        extension = f".{extension}" if extension else os.path.splitext(self.file.path)[1]
        return f"{slugify(self.invoice_number)}_{slugify(self.project_details.name)}{extension}"

    def filename_for_zip(self, extension=None):
        extension = f".{extension}" if extension else os.path.splitext(self.file.path)[1]
        return export_invoice_filename(self) + extension

    def render_and_save_file(self):
        from NEMO_billing.invoices.renderers import invoice_renderer_class

        content = invoice_renderer_class.render_invoice(self)
        content.seek(0)
        self.file = ContentFile(content.read(), "invoice." + invoice_renderer_class.get_file_extension())
        content.close()
        self.save(update_fields=["file"])

    def _email_invoice(self, template_name) -> bool:
        attachment = create_email_attachment(self.file, self.filename())
        sent = render_and_send_email(
            template_name,
            {"invoice": self},
            to=self.project_details.email_to(),
            from_email=self.configuration.email_from,
            cc=self.configuration.email_cc,
            attachments=[attachment],
        )
        return bool(sent)

    def send(self) -> bool:
        # Email the invoice. If it was already sent, don't change the due date
        if not self.voided_date and self.reviewed_date:
            self.last_sent_date = timezone.now()
            if not self.sent_date:
                self.sent_date = timezone.now()
                self.due_date = timezone.now() + timedelta(days=self.configuration.invoice_due_in)
            if self._email_invoice("invoices/email/email_send_invoice"):
                self.save()
                return True
        return False

    def send_reminder(self) -> bool:
        if self.sent_date:
            if self.total_payments_received() < self.total_amount:
                # Invoice hasn't been paid in full, reminder should be sent
                self.last_reminder_sent_date = timezone.now()
                if self._email_invoice("invoices/email/email_send_invoice_reminder"):
                    self.save()
                    return True
        return False

    def send_voided_email(self) -> bool:
        # Notify folks that the invoice was voided
        from NEMO_billing.invoices.customization import InvoiceCustomization

        # Only send the voided email if the invoice has already been sent
        if self.sent_date and InvoiceCustomization.get_bool("invoice_send_email_when_voided"):
            render_and_send_email(
                "invoices/email/email_send_invoice_voided",
                {"invoice": self},
                to=self.project_details.email_to(),
                from_email=self.configuration.email_from,
                cc=self.configuration.email_cc,
            )
            return True
        return False

    def sorted_core_facilities(self, detail_items=None) -> List[str]:
        if detail_items:
            core_facilities = list({item.core_facility for item in detail_items})
        else:
            core_facilities = list(self.invoicedetailitem_set.values_list("core_facility", flat=True).distinct())
        core_facilities.sort(key=lambda x: x if x else "", reverse=True)
        return core_facilities

    def summary_dict(self) -> Dict[str, List]:
        summary_details = {}
        for core_facility in self.sorted_core_facilities():
            core_facility_details = self.invoicesummaryitem_set.filter(core_facility=core_facility)
            if self.configuration.hide_zero_charge:
                core_facility_details = core_facility_details.exclude(amount=0)
            summary_details.setdefault(core_facility, core_facility_details)
        details = self.invoicesummaryitem_set.filter(core_facility=None)
        if self.configuration.hide_zero_charge:
            details = details.exclude(amount=0)
        summary_details.setdefault(None, details)
        return summary_details

    def details_dict(self) -> Dict[str, Dict[str, List]]:
        details = {}
        for core_facility in self.sorted_core_facilities():
            details.setdefault(core_facility, {})
            core_facility_items = details.get(core_facility)
            core_facility_items.setdefault("tool_usage", self.tool_usage_details(core_facility=core_facility))
            core_facility_items.setdefault("area_access", self.area_access_details(core_facility=core_facility))
            core_facility_items.setdefault("staff_charges", self.staff_charge_details(core_facility=core_facility))
            core_facility_items.setdefault(
                "consumable_withdrawals", self.consumable_withdrawal_details(core_facility=core_facility)
            )
            core_facility_items.setdefault("trainings", self.training_details(core_facility=core_facility))
            core_facility_items.setdefault(
                "missed_reservations", self.missed_reservation_details(core_facility=core_facility)
            )
            core_facility_items.setdefault("custom_charges", self.custom_charges_details(core_facility=core_facility))
        return details

    def invoice_details(self):
        details = self.invoicedetailitem_set.all()
        if self.configuration.hide_zero_charge:
            details = details.exclude(amount=0)
        return details

    def tool_usage_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.TOOL_USAGE.value)
            .order_by("start")
        )

    def area_access_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.AREA_ACCESS.value)
            .order_by("start")
        )

    def staff_charge_details(self, core_facility: str):
        return self.invoice_details().filter(core_facility=core_facility, item_type=BillableItemType.STAFF_CHARGE.value)

    def consumable_withdrawal_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.CONSUMABLE.value)
            .order_by("start")
        )

    def training_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.TRAINING.value)
            .order_by("start")
        )

    def missed_reservation_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.MISSED_RESERVATION.value)
            .order_by("start")
        )

    def custom_charges_details(self, core_facility: str):
        return (
            self.invoice_details()
            .filter(core_facility=core_facility, item_type=BillableItemType.CUSTOM_CHARGE.value)
            .order_by("start")
        )

    def total_amount_display(self) -> str:
        return display_amount(self.total_amount, self.configuration)

    def total_payments_received(self) -> Decimal:
        return self.invoicepayment_set.aggregate(
            total_received=Coalesce(
                Sum(Case(When(payment_received__isnull=False, then="amount"), default=Decimal(0))), Decimal(0)
            )
        )["total_received"]

    def total_payments_processed(self) -> Decimal:
        return self.invoicepayment_set.aggregate(
            total_processed=Coalesce(
                Sum(Case(When(payment_processed__isnull=False, then="amount"), default=Decimal(0))), Decimal(0)
            )
        )["total_processed"]

    def total_outstanding_amount(self) -> Decimal:
        return self.total_amount - self.total_payments_received()

    def total_outstanding_display(self) -> str:
        return display_amount(self.total_outstanding_amount(), self.configuration)

    def total_payments_display(self) -> str:
        pending_display = f" ({display_amount(self.total_payments_received()-self.total_payments_processed(), self.configuration)} pending)"
        return f"{self.total_outstanding_display()}{pending_display}"

    def mark_as_paid_in_full(self, user):
        outstanding = self.total_outstanding_amount()
        if outstanding > 0:
            self.record_payment(user, outstanding, timezone.now(), timezone.now())

    def record_payment(self, user: User, amount: Decimal, received=None, processed=None, note=None) -> InvoicePayment:
        return InvoicePayment.objects.create(
            invoice=self,
            created_by=user,
            updated_by=user,
            payment_received=received,
            payment_processed=processed,
            amount=amount,
            note=note,
        )

    def tax_display(self) -> str:
        return display_amount(
            self.invoicesummaryitem_set.aggregate(
                total_tax=Coalesce(
                    Sum(
                        Case(
                            When(summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.TAX, then="amount"),
                            default=Decimal(0),
                        )
                    ),
                    Decimal(0),
                )
            )["total_tax"],
            self.configuration,
        )

    def __str__(self):
        created_date = f" ({self.created_date.date()})" if self.created_date else ""
        return f"{self.invoice_number}: {self.project_details.name}{created_date}"

    class Meta:
        ordering = ["-created_date", "-invoice_number"]


class InvoiceDetailItem(BaseModel):
    content_type = models.ForeignKey(ContentType, null=True, blank=True, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey("content_type", "object_id")
    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE)
    core_facility = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_MEDIUM_LENGTH)
    item_type = models.IntegerField(choices=BillableItemType.choices())
    name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)
    quantity = models.DecimalField(decimal_places=2, max_digits=8)
    start = models.DateTimeField()
    end = models.DateTimeField()
    user = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)
    rate = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_SMALL_LENGTH)
    amount = models.DecimalField(decimal_places=2, max_digits=14)
    discount = models.DecimalField(null=True, blank=True, decimal_places=2, max_digits=14)
    waived = models.BooleanField(default=False)

    def quantity_display(self):
        if self.item_type and BillableItemType(self.item_type).is_time_type():
            return f"{self.quantity:.2f} min"
        else:
            return f"{self.quantity:.2f}"

    def amount_display(self):
        amount = display_amount(self.amount, self.invoice.configuration)
        return f"{amount}{' (waived)' if self.waived else ''}"

    class Meta:
        indexes = [
            models.Index(fields=["content_type", "object_id"]),
        ]


class InvoiceSummaryItem(BaseModel):
    class InvoiceSummaryItemType(object):
        ITEM = 1  # Any billable summary item
        SUBTOTAL = 2  # Facility subtotal (sum of billable summary item minus discounts)
        DISCOUNT_SUBTOTAL = 3  # Facility amount for discount (only used to calculate future accumulation discounts)
        DISCOUNT = 4  # Facility discount
        TAX = 5  # Tax
        OTHER = 6  # Other
        FUND = 7  # Other
        choices = (
            (ITEM, "item"),
            (SUBTOTAL, "sub_total"),
            (DISCOUNT_SUBTOTAL, "discount_sub_total"),
            (DISCOUNT, "discount"),
            (TAX, "tax"),
            (OTHER, "other"),
            (FUND, "fund"),
        )

    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE)
    summary_item_type = models.IntegerField(choices=InvoiceSummaryItemType.choices)
    item_type = models.IntegerField(null=True, blank=True, choices=BillableItemType.choices())
    core_facility = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_MEDIUM_LENGTH)
    name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)
    details = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_SMALL_LENGTH)
    amount = models.DecimalField(null=True, blank=True, decimal_places=2, max_digits=14)

    def amount_display(self):
        return display_amount(self.amount, self.invoice.configuration)

    def category_name_for_item_type(self):
        return category_name_for_item_type(self.item_type)


class InvoicePayment(BaseModel):
    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE)
    payment_received = models.DateField(help_text="Date when payment was received")
    payment_processed = models.DateField(null=True, blank=True, help_text="Date when payment was processed")
    amount = models.DecimalField(decimal_places=2, max_digits=14, help_text="Amount received")
    note = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_MEDIUM_LENGTH, help_text="Payment note")
    created_date = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, related_name="payment_created_by_set", on_delete=models.PROTECT)
    updated_date = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(User, related_name="payment_updated_by_set", on_delete=models.PROTECT)

    def amount_display(self):
        return display_amount(self.amount, self.invoice.configuration)

    def __str__(self):
        return f"Payment for invoice {self.invoice.invoice_number}"

    class Meta:
        ordering = ["-payment_received"]



import datetime
from typing import Set

from NEMO.decorators import accounting_or_user_office_or_manager_required
from NEMO.typing import QuerySetType
from NEMO.utilities import extract_optional_beginning_and_end_dates, get_month_timeframe, month_list
from NEMO.views.pagination import SortedPaginator
from NEMO.views.usage import get_managed_projects
from dateutil.relativedelta import relativedelta
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from NEMO_billing.models import CoreFacility
from NEMO_billing.prepayments.models import Fund, ProjectPrepaymentDetail
from NEMO_billing.utilities import get_charges_amount_between, number_of_months_between_dates


@login_required
def usage_project_prepayments(request):
    start_date, end_date = get_month_timeframe()
    projects = get_managed_projects(request.user)
    prepaid_projects = ProjectPrepaymentDetail.objects.filter(
        project__in=[project for project in projects if project.active]
    )
    add_fund_info_to_prepaid_projects(request, prepaid_projects, start_date)
    return render(request, "prepayments/prepaid_project_status_table.html", {"page": prepaid_projects})


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def prepaid_project_status(request):
    # Default is year to date
    start_date, end_date = get_month_timeframe()
    start_date = start_date.replace(month=1)
    if request.GET.get("start"):
        submitted_start_date, submitted_end_date = extract_optional_beginning_and_end_dates(request.GET, date_only=True)
        if submitted_start_date:
            start_date = start_date.replace(year=submitted_start_date.year, month=submitted_start_date.month)
    page = SortedPaginator(ProjectPrepaymentDetail.objects.all(), request, order_by="project").get_current_page()
    add_fund_info_to_prepaid_projects(request, page.object_list, start_date)
    dictionary = {
        "start": start_date,
        "month_list": month_list(),
        "page": page,
        "core_facilities_exist": CoreFacility.objects.exists(),
    }
    return render(request, "prepayments/prepaid_project_status.html", dictionary)


def add_fund_info_to_prepaid_projects(request, prepaid_projects: QuerySetType[ProjectPrepaymentDetail], start_date):
    for prepaid_project in prepaid_projects:
        try:
            charges, charges_amount, funds = get_prepayment_info_for_status(prepaid_project, start_date)
            prepaid_project.total_funds = sum(fund.amount for fund in funds)
            prepaid_project.balance = sum(fund.balance for fund in funds)
            prepaid_project.charges_amount = sum(charge.amount for charge in charges)
            prepaid_project.taxes_amount = (
                prepaid_project.charges_amount * prepaid_project.configuration.tax_amount()
                if not prepaid_project.project.projectbillingdetails.no_tax
                else 0
            )
            prepaid_project.total_charges_amount = prepaid_project.charges_amount + prepaid_project.taxes_amount
            prepaid_project.total_funds_left = prepaid_project.balance - prepaid_project.total_charges_amount
            prepaid_project.charges = charges
            prepaid_project.charges.reverse()
        except Exception as e:
            messages.error(request, str(e), extra_tags="data-trigger=manual")


def get_prepayment_info_for_status(prepaid_project: ProjectPrepaymentDetail, fund_start_date):
    # Returns total charges, total charges amount since last updated, and all active funds since fund_start_date
    until = datetime.date.today()
    # Balance last updated is last day of previous month, so start first day of next month
    start_in_month = (prepaid_project.balance_last_updated + relativedelta(months=1)).replace(day=1)
    charges, charges_amount = get_charges_amount_between(
        prepaid_project.project, prepaid_project.configuration, start_in_month, until
    )
    # Check funds month to month
    funds: Set[Fund] = set()
    months = number_of_months_between_dates(until, fund_start_date)
    for month in range(0, months + 1):
        month_date = fund_start_date + relativedelta(months=month)
        # beginning and end of the month
        start, end = get_month_timeframe(month_date.isoformat())
        # only need to check funds valid at this date (expired or inactive won't be returned by active_funds)
        try:
            new_funds: QuerySetType[Fund] = prepaid_project.active_funds(end.date(), include_zero_balance=True)
            funds.update(new_funds)
        except Exception:
            pass
    charges.sort(key=lambda x: x.start)
    return charges, charges_amount, funds



from NEMO.serializers import ModelSerializer
from NEMO.views.api import ModelViewSet, key_filters
from rest_flex_fields import FlexFieldsModelSerializer

from NEMO_billing.prepayments.models import Fund, FundType, ProjectPrepaymentDetail


class ProjectPrepaymentDetailSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = ProjectPrepaymentDetail
        fields = "__all__"
        expandable_fields = {
            "project": "NEMO.serializers.ProjectSerializer",
        }


class FundTypeSerializer(ModelSerializer):
    class Meta:
        model = FundType
        fields = "__all__"


class FundSerializer(FlexFieldsModelSerializer, ModelSerializer):
    class Meta:
        model = Fund
        fields = "__all__"
        expandable_fields = {
            "fund_type": "NEMO_billing.prepayments.api.FundTypeSerializer",
            "project_prepayment": "NEMO_billing.prepayments.api.ProjectPrepaymentDetailSerializer",
        }


class FundTypeViewSet(ModelViewSet):
    filename = "fund_types"
    queryset = FundType.objects.all()
    serializer_class = FundTypeSerializer


class FundViewSet(ModelViewSet):
    filename = "funds"
    queryset = Fund.objects.all()
    serializer_class = FundSerializer
    filterset_fields = {
        "id": key_filters,
        "fund_type": key_filters,
        "project_prepayment": key_filters,
    }


class ProjectPrepaymentDetailViewSet(ModelViewSet):
    filename = "project_prepayments"
    queryset = ProjectPrepaymentDetail.objects.all()
    serializer_class = ProjectPrepaymentDetailSerializer
    filterset_fields = {
        "id": key_filters,
        "project": key_filters,
    }



# Generated by Django 3.2.19 on 2023-06-26 20:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0004_version_2_4_0"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="institution",
            options={"ordering": ["name"]},
        ),
    ]



TOOL_CORE_FACILITY_REQUIRED = False
AREA_CORE_FACILITY_REQUIRED = False
CONSUMABLE_CORE_FACILITY_REQUIRED = False
STAFF_CHARGE_CORE_FACILITY_REQUIRED = False
CUSTOM_CHARGE_CORE_FACILITY_REQUIRED = False

STAFF_TOOL_USAGE_AS_STAFF_CHARGE = True
STAFF_AREA_ACCESS_AS_STAFF_CHARGE = True

NEMO_POLICY_CLASS = "NEMO_billing.policy.BillingPolicy"



from NEMO.utilities import format_datetime
from django import forms
from django.contrib import admin
from django.contrib.admin import widgets

from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.prepayments.models import Fund, FundType, ProjectPrepaymentDetail
from NEMO_billing.utilities import IntMultipleChoiceField


class ProjectPrepaymentDetailAdminForm(forms.ModelForm):
    charge_types = IntMultipleChoiceField(
        choices=BillableItemType.choices_except(BillableItemType.CUSTOM_CHARGE, BillableItemType.CONSUMABLE),
        required=True,
        widget=widgets.FilteredSelectMultiple(verbose_name="Types", is_stacked=False),
    )

    class Meta:
        model = ProjectPrepaymentDetail
        fields = "__all__"


@admin.register(Fund)
class FundAdmin(admin.ModelAdmin):
    list_display = [
        "__str__",
        "get_project",
        "get_account",
        "fund_type",
        "amount",
        "balance",
        "get_balance_date_display",
        "get_start_display",
        "get_expiration_display",
        "reference",
        "balance_warning_percent",
    ]
    list_filter = [
        ("fund_type", admin.RelatedOnlyFieldListFilter),
        ("project_prepayment__project", admin.RelatedOnlyFieldListFilter),
        ("project_prepayment__project__account", admin.RelatedOnlyFieldListFilter),
    ]

    @admin.display(description="Project", ordering="project_prepayment__project")
    def get_project(self, obj: Fund):
        return obj.project_prepayment.project

    @admin.display(description="Account", ordering="project_prepayment__project__account")
    def get_account(self, obj: Fund):
        return obj.project_prepayment.project.account

    @admin.display(description="Start", ordering=["start_year", "start_month"])
    def get_start_display(self, obj: Fund):
        return format_datetime(obj.start_date, "F Y")

    @admin.display(description="Expires", ordering=["expiration_year", "expiration_month"])
    def get_expiration_display(self, obj: Fund):
        if obj.expiration_date:
            return format_datetime(obj.expiration_date, "F Y")

    @admin.display(description="Balance date", ordering="balance_date")
    def get_balance_date_display(self, obj: Fund):
        return format_datetime(obj.project_prepayment.balance_last_updated, "SHORT_DATE_FORMAT")


@admin.register(ProjectPrepaymentDetail)
class ProjectPrepaymentDetailAdmin(admin.ModelAdmin):
    list_display = ["project", "get_charge_types", "get_only_core_facilities", "overdraft_amount_allowed"]
    filter_horizontal = ["only_core_facilities"]
    form = ProjectPrepaymentDetailAdminForm

    @admin.display(description="Core facilities allowed")
    def get_only_core_facilities(self, instance: ProjectPrepaymentDetail):
        return instance.get_only_core_facilities_display()

    @admin.display(description="Charge types allowed")
    def get_charge_types(self, instance: ProjectPrepaymentDetail):
        return instance.get_charge_types_display()


admin.site.register(FundType)



# Generated by Django 3.2.19 on 2023-06-02 00:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0004_version_2_0_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="rate",
            name="daily",
            field=models.BooleanField(
                default=False, help_text="Check this box to only charge once per day for this item"
            ),
        ),
    ]



from NEMO.urls import router, sort_urls
from django.urls import path, re_path

from NEMO_billing import api, views

# Rest API URLs
router.register(r"billing/core_facilities", api.CoreFacilityViewSet)
router.register(r"billing/custom_charges", api.CustomChargeViewSet)
router.register(r"billing/departments", api.DepartmentViewSet)
router.register(r"billing/institution_types", api.InstitutionTypeViewSet)
router.register(r"billing/institutions", api.InstitutionViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("custom_charges/", views.custom_charges, name="custom_charges"),
    path("custom_charge/", views.create_or_modify_custom_charge, name="create_custom_charge"),
    path("custom_charge/<int:custom_charge_id>", views.create_or_modify_custom_charge, name="edit_custom_charge"),
    path(
        "get_projects_for_custom_charges/",
        views.get_projects_for_custom_charges,
        name="get_projects_for_custom_charges",
    ),
    # Overriding NEMO create staff charge URLs to add core facility selection
    path("staff_charges/", views.custom_staff_charges, name="staff_charges"),
    path("begin_staff_charge/", views.custom_begin_staff_charge, name="begin_staff_charge"),
    # Overriding NEMO enable tool URL to set the core facility on staff charge based on the tool's core facility
    path(
        "enable_tool/<int:tool_id>/user/<int:user_id>/project/<int:project_id>/staff_charge/<str:staff_charge>/",
        views.custom_enable_tool,
        name="enable_tool",
    ),
    # Override broadcast email to add new search for PIs by institution types
    path("email_broadcast/", views.email_broadcast, name="email_broadcast"),
    re_path(
        r"^email_broadcast/(?P<audience>tool|area|account|project|project-pis|user|tool-reservation|group|pi_institution_types)/$",
        views.email_broadcast,
        name="email_broadcast",
    ),
]



# Generated by Django 2.2.13 on 2021-03-13 12:49

import NEMO.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import NEMO_billing.invoices.utilities


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("rates", "0001_initial"),
        ("NEMO", "0028_version_3_9_0"),
    ]

    def create_invoice_customizations(apps, schema_editor):
        CustomizationModel = apps.get_model("NEMO", "Customization")
        CustomizationModel.objects.update_or_create(name="invoice_number_format", defaults={"value": "{:04d}"})
        CustomizationModel.objects.update_or_create(name="invoice_number_current", defaults={"value": "0"})

    operations = [
        migrations.CreateModel(
            name="InvoiceConfiguration",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "name",
                    models.CharField(help_text="The name of this invoice configuration", max_length=200, unique=True),
                ),
                (
                    "invoice_due_in",
                    models.PositiveIntegerField(
                        default=30, help_text="The default number of days invoices are due after"
                    ),
                ),
                (
                    "reminder_frequency",
                    models.PositiveIntegerField(
                        null=True,
                        blank=True,
                        default=7,
                        help_text="How often to send a reminder. Default value is 7, meaning every week after past due invoice",
                    ),
                ),
                ("email_from", models.EmailField(help_text="The email address used to send invoices and reminders")),
                (
                    "email_cc",
                    NEMO.fields.MultiEmailField(
                        null=True,
                        blank=True,
                        help_text="Email to cc the invoice to. A comma-separated list can be used",
                        max_length=2000,
                    ),
                ),
                (
                    "terms",
                    models.TextField(
                        null=True, blank=True, help_text="Terms and conditions to be included in the invoice"
                    ),
                ),
                ("merchant_name", models.CharField(max_length=255)),
                (
                    "merchant_details",
                    models.TextField(
                        blank=True,
                        help_text="The merchant details to be included in the invoice (address, phone number etc.)",
                        null=True,
                    ),
                ),
                (
                    "merchant_logo",
                    models.ImageField(
                        blank=True, null=True, upload_to=NEMO_billing.invoices.utilities.get_merchant_logo_filename
                    ),
                ),
                ("currency", models.CharField(default="USD", max_length=4)),
                ("currency_symbol", models.CharField(null=True, blank=True, max_length=4, default="$")),
                (
                    "tax",
                    models.DecimalField(
                        null=True,
                        blank=True,
                        decimal_places=3,
                        max_digits=5,
                        help_text="Tax in percent. For 20.5% enter 20.5",
                    ),
                ),
                ("tax_name", models.CharField(max_length=50, null=True, blank=True, default="VAT")),
                (
                    "detailed_invoice",
                    models.BooleanField(
                        default=True, help_text="Check this box if customers should receive a detailed invoice."
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ProjectBillingDetails",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "category",
                    models.ForeignKey(null=True, blank=True, on_delete=models.SET_NULL, to="rates.RateCategory"),
                ),
                (
                    "project_name",
                    models.CharField(
                        blank=True,
                        help_text="The project name that will appear on the invoices. Leave blank to use NEMO project name",
                        max_length=200,
                        null=True,
                    ),
                ),
                (
                    "contact_name",
                    models.CharField(
                        blank=True, help_text="The contact name to use in the invoice email", max_length=255, null=True
                    ),
                ),
                (
                    "contact_email",
                    NEMO.fields.MultiEmailField(
                        null=True,
                        blank=True,
                        help_text="Email to send the invoice to. A comma-separated list can be used. Leave blank to use project managers/PIs emails",
                        max_length=2000,
                    ),
                ),
                (
                    "details",
                    models.TextField(
                        blank=True,
                        help_text="The project details to be included in the invoice (address, etc.)",
                        null=True,
                    ),
                ),
                (
                    "no_charge",
                    models.BooleanField(
                        default=False, help_text="Check this box if invoices should not be created for this project."
                    ),
                ),
                ("project", models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to="NEMO.Project")),
            ],
            options={"verbose_name_plural": "Project details"},
        ),
        migrations.CreateModel(
            name="Invoice",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "invoice_number",
                    models.CharField(
                        blank=True, help_text="Leave blank to be assigned automatically", max_length=100, unique=True
                    ),
                ),
                ("start", models.DateTimeField()),
                ("end", models.DateTimeField()),
                ("due_date", models.DateField(null=True, blank=True)),
                ("sent_date", models.DateTimeField(null=True, blank=True)),
                ("last_sent_date", models.DateTimeField(null=True, blank=True)),
                ("last_reminder_sent_date", models.DateTimeField(null=True, blank=True)),
                ("reviewed_date", models.DateTimeField(null=True, blank=True)),
                ("created_date", models.DateTimeField(auto_now_add=True)),
                ("voided_date", models.DateTimeField(null=True, blank=True)),
                ("total_amount", models.DecimalField(decimal_places=2, max_digits=14)),
                (
                    "configuration",
                    models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to="invoices.InvoiceConfiguration"),
                ),
                (
                    "reviewed_by",
                    models.ForeignKey(
                        null=True,
                        blank=True,
                        related_name="reviewed_invoice_set",
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "voided_by",
                    models.ForeignKey(
                        null=True,
                        blank=True,
                        related_name="voided_invoice_set",
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL),
                ),
                (
                    "file",
                    models.FileField(
                        null=True, blank=True, upload_to=NEMO_billing.invoices.utilities.get_invoice_document_filename
                    ),
                ),
                (
                    "project_details",
                    models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to="invoices.ProjectBillingDetails"),
                ),
            ],
            options={
                "ordering": ["-created_date", "-invoice_number"],
            },
        ),
        migrations.CreateModel(
            name="InvoicePayment",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("payment_received", models.DateField(help_text="Date when payment was received")),
                (
                    "payment_processed",
                    models.DateField(blank=True, help_text="Date when payment was processed", null=True),
                ),
                ("amount", models.DecimalField(decimal_places=2, max_digits=14, help_text="Amount received")),
                ("created_date", models.DateTimeField(auto_now_add=True)),
                ("updated_date", models.DateTimeField(auto_now=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="payment_created_by_set",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("invoice", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="invoices.Invoice")),
                (
                    "updated_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="payment_updated_by_set",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ["-payment_received"],
            },
        ),
        migrations.CreateModel(
            name="InvoiceSummaryItem",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "summary_item_type",
                    models.IntegerField(
                        choices=[
                            (1, "item"),
                            (2, "category"),
                            (3, "small_category"),
                            (4, "sub_total"),
                            (5, "discount"),
                            (6, "facility_discount"),
                            (7, "tax"),
                            (8, "other"),
                        ]
                    ),
                ),
                ("core_facility", models.CharField(blank=True, max_length=255, null=True)),
                ("name", models.CharField(max_length=255)),
                ("details", models.CharField(blank=True, max_length=100, null=True)),
                ("amount", models.DecimalField(null=True, blank=True, decimal_places=2, max_digits=14)),
                ("invoice", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="invoices.Invoice")),
            ],
        ),
        migrations.CreateModel(
            name="InvoiceDetailItem",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("core_facility", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "item_type",
                    models.IntegerField(
                        choices=[
                            (1, "tool_usage"),
                            (2, "area_access"),
                            (3, "consumable"),
                            (4, "missed_reservation"),
                            (5, "staff_charge"),
                            (6, "training_session"),
                            (7, "custom_charge"),
                        ]
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("quantity", models.DecimalField(decimal_places=2, max_digits=8)),
                ("start", models.DateTimeField()),
                ("end", models.DateTimeField()),
                ("user", models.CharField(max_length=200)),
                ("rate", models.CharField(blank=True, max_length=100, null=True)),
                ("amount", models.DecimalField(decimal_places=2, max_digits=14)),
                ("invoice", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="invoices.Invoice")),
            ],
        ),
        migrations.RunPython(create_invoice_customizations),
    ]



from django.urls import path

from NEMO_billing.cap_discount import views

urlpatterns = [
    path("usage_cap_discounts/", views.usage_cap_discounts, name="usage_cap_discounts"),
    path("usage_cap_discounts_user/<int:user_id>/", views.usage_cap_discounts_user, name="usage_cap_discounts_user"),
    path(
        "usage_cap_discounts_account/<int:account_id>/",
        views.usage_cap_discounts_account,
        name="usage_cap_discounts_account",
    ),
    path("cap_discount_status/", views.cap_discount_status, name="cap_discount_status"),
    path(
        "cap_discount_status/configuration/<int:configuration_id>/",
        views.cap_discount_status,
        name="cap_discount_status_configuration",
    ),
]



import io
import zipfile
from datetime import datetime
from decimal import Decimal
from logging import getLogger
from typing import List

from NEMO.decorators import accounting_or_manager_required, accounting_or_user_office_or_manager_required, synchronized
from NEMO.models import Account
from NEMO.utilities import (
    BasicDisplayTable,
    date_input_format,
    export_format_datetime,
    format_datetime,
    month_list,
    queryset_search_filter,
)
from NEMO.views.customization import ProjectsAccountsCustomization
from NEMO.views.pagination import SortedPaginator
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Case, F, Func, OuterRef, Prefetch, Subquery, Sum, When
from django.db.models.functions import Coalesce, Length
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import date_format
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO_billing.exceptions import BillingException
from NEMO_billing.invoices.customization import InvoiceCustomization
from NEMO_billing.invoices.exceptions import (
    InvoiceAlreadyExistException,
    InvoiceGenerationException,
    InvoiceItemsNotInFacilityException,
    NoProjectCategorySetException,
    NoProjectDetailsSetException,
    NoRateSetException,
)
from NEMO_billing.invoices.models import (
    BillableItemType,
    Invoice,
    InvoiceConfiguration,
    InvoicePayment,
    InvoiceSummaryItem,
)
from NEMO_billing.invoices.processors import invoice_data_processor_class as processor
from NEMO_billing.invoices.renderers import CSVInvoiceRenderer
from NEMO_billing.invoices.utilities import category_name_for_item_type, display_amount
from NEMO_billing.models import CoreFacility
from NEMO_billing.rates.models import RateType

invoices_logger = getLogger(__name__)


@accounting_or_user_office_or_manager_required
@require_GET
def invoices(request, year: int = None, month: int = None):
    invoice_list = Invoice.objects.filter(voided_date=None)
    if year:
        invoice_list = invoice_list.filter(start__year=year)
        if month:
            invoice_list = invoice_list.filter(start__month=month)
    # Add outstanding balance (sortable column)
    outstanding_sub = (
        InvoicePayment.objects.filter(invoice=OuterRef("pk"))
        .annotate(total_payments=Coalesce(Func("amount", function="Sum"), Decimal(0)))
        .values("total_payments")
        .order_by()
    )
    invoice_list = invoice_list.annotate(outstanding=F("total_amount") - Subquery(outstanding_sub))

    # Add total tax (sortable column). TODO: might need to change to subquery to avoid issue with duplicates in list of related fields
    invoice_list = invoice_list.annotate(
        total_tax=Sum(
            Case(
                When(
                    invoicesummaryitem__summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.TAX,
                    then=F("invoicesummaryitem__amount"),
                ),
                default=Decimal(0),
            )
        )
    )

    core_facilities = CoreFacility.objects.all()

    for facility in core_facilities:
        sub = InvoiceSummaryItem.objects.filter(
            invoice=OuterRef("pk"),
            core_facility=facility.name,
            summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL,
        ).values("amount")
        invoice_list = invoice_list.annotate(**{f"facility_total_{facility.id}": Subquery(sub)})
    # Add general (None) subtotal too
    sub = InvoiceSummaryItem.objects.filter(
        invoice=OuterRef("pk"), core_facility=None, summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
    ).values("amount")
    invoice_list = invoice_list.annotate(**{f"facility_total_general": Subquery(sub)})

    # Prefetch facility subtotals, projects and configurations
    invoice_list = (
        invoice_list.prefetch_related(
            Prefetch(
                "invoicesummaryitem_set",
                queryset=InvoiceSummaryItem.objects.filter(
                    summary_item_type=InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
                ),
                to_attr="facility_subtotals",
            )
        )
        .prefetch_related("project_details__project")
        .prefetch_related("configuration")
    )

    # Fixed sort by number in case they are longer
    paginator = SortedPaginator(invoice_list, request, order_by="-created_date")
    order = request.GET.get("o")
    if order == "-invoice_number":
        paginator.object_list = paginator.object_list.order_by(Length("invoice_number").desc(), "-invoice_number")
    elif order == "invoice_number":
        paginator.object_list = paginator.object_list.order_by(Length("invoice_number").asc(), "invoice_number")

    page = paginator.get_current_page()

    display_general_facility = not core_facilities.exists() or not settings.INVOICE_ALL_ITEMS_MUST_BE_IN_FACILITY

    csv_export = bool(request.GET.get("csv", False))
    if csv_export:
        table_result = BasicDisplayTable()
        table_result.headers = [
            ("number", "Number"),
            ("created", "Created"),
            ("month", "Month"),
            ("project", "Project"),
            ("identifier", ProjectsAccountsCustomization.get("project_application_identifier_name")),
            ("account", "Account"),
            ("sent", "Sent"),
            ("due", "Due"),
            ("reviewed", "Reviewed"),
            ("reviewed_by", "Reviewed by"),
            ("total", "Total"),
        ]
        for core_facility in core_facilities:
            table_result.add_header(("facility_" + core_facility.name, core_facility.name))
        if display_general_facility:
            table_result.add_header(("facility_", "General"))
        table_result.add_header(("tax", "Tax"))
        table_result.add_header(("outstanding", "Outstanding"))
        table_result.add_header(("emails", "Emails"))
        for invoice in page:
            row = {
                "number": invoice.invoice_number,
                "created": format_datetime(invoice.created_date, "SHORT_DATETIME_FORMAT"),
                "month": format_datetime(invoice.start, "F Y"),
                "project": invoice.project_details.name,
                "identifier": invoice.project_details.project.application_identifier,
                "account": invoice.project_details.project.account.name,
                "sent": format_datetime(invoice.last_sent_date, "SHORT_DATE_FORMAT") if invoice.last_sent_date else "",
                "due": format_datetime(invoice.due_date, "SHORT_DATE_FORMAT") if invoice.due_date else "",
                "reviewed": (
                    format_datetime(invoice.reviewed_date, "SHORT_DATE_FORMAT") if invoice.reviewed_date else ""
                ),
                "reviewed_by": invoice.reviewed_by.get_name() if invoice.reviewed_by else "",
                "total": invoice.total_amount_display(),
                "outstanding": display_amount(invoice.outstanding, invoice.configuration),
                "tax": display_amount(invoice.total_tax, invoice.configuration),
                "emails": ",".join(invoice.project_details.email_to()),
            }
            facilities_subtotals = {sub.core_facility: sub for sub in invoice.facility_subtotals}
            for core_facility in core_facilities:
                row["facility_" + core_facility.name] = display_amount(
                    (
                        facilities_subtotals[core_facility.name].amount
                        if core_facility.name in facilities_subtotals
                        else 0
                    ),
                    invoice.configuration,
                )
            if display_general_facility:
                row["facility_"] = display_amount(
                    facilities_subtotals[None].amount if None in facilities_subtotals else 0, invoice.configuration
                )
            table_result.add_row(row)
        response = table_result.to_csv()
        filename = f"invoices_{export_format_datetime()}.csv"
        response["Content-Disposition"] = f'attachment; filename="{filename}"'
        return response

    request.session["current_invoice_list_id"] = [invoice.id for invoice in page]

    month_list_since = InvoiceCustomization.get_date("invoice_month_list_since", False) or datetime(
        year=2021, month=1, day=1
    )

    return render(
        request,
        "invoices/invoices.html",
        {
            "page": page,
            "month_list": month_list(since=month_list_since),
            "accounts": Account.objects.all(),
            "configuration_list": InvoiceConfiguration.objects.all(),
            "filter_month_list": Invoice.objects.filter(voided_date=None).dates("start", "month", order="DESC"),
            "filter_year": year or datetime.now().year,
            "filter_month": month,
            "core_facilities": core_facilities,
            "display_general_facility": display_general_facility,
        },
    )


@accounting_or_manager_required
@require_POST
@synchronized()
def generate_monthly_invoices(request):
    extra_tags = "data-speed=20000"
    warning_extra_tags = "data-speed=40000"
    error_extra_tags = "data-speed=40000 data-trigger=manual"
    generated_invoices = []
    has_errors = True
    try:
        account_id: str = request.POST["account_id"]
        configuration_id = request.POST["configuration_id"]
        configuration = get_object_or_404(InvoiceConfiguration, id=configuration_id)
        month = request.POST["month"]
        if account_id == "All":
            for account in Account.objects.all():
                invoice_list = processor.generate_invoice_for_account(month, account, configuration, request.user)
                generated_invoices.extend(invoice_list)
        else:
            account: Account = get_object_or_404(Account, id=account_id)
            invoice_list = processor.generate_invoice_for_account(month, account, configuration, request.user)
            generated_invoices.extend(invoice_list)
        has_errors = False
    except NoProjectDetailsSetException as e:
        link = reverse("project", args=[e.project.id])
        message = "Invoice generation failed: " + e.msg + f" - click <a href='{link}'>here</a> to add some."
        messages.error(request, mark_safe(message), error_extra_tags)
    except NoRateSetException as e:
        link = create_rate_link(e.rate_type, e.tool, e.area, e.consumable)
        message = "Invoice generation failed: " + e.msg + f" - click <a href='{link}'>here</a> to create one."
        messages.error(request, mark_safe(message), error_extra_tags)
    except InvoiceAlreadyExistException as e:
        link = reverse("view_invoice", args=[e.invoice.id])
        message = "Invoice generation failed: " + e.msg + f" - click <a href='{link}'>here</a> to view it."
        messages.error(request, mark_safe(message), error_extra_tags)
    except NoProjectCategorySetException as e:
        link = reverse("project", args=[e.project.id])
        message = "Invoice generation failed: " + e.msg + f" - click <a href='{link}'>here</a> to set it."
        messages.error(request, mark_safe(message), error_extra_tags)
    except InvoiceItemsNotInFacilityException as e:
        messages.error(request, e.msg, error_extra_tags)
    except InvoiceGenerationException as e:
        link = reverse("re_render_invoice", args=[e.invoice.id])
        message = e.msg + f" - click <a href='{link}'>here</a> to try generating it again."
        messages.error(request, mark_safe(message), error_extra_tags)
    except BillingException as e:
        messages.error(request, str(e), error_extra_tags)
    except Exception as e:
        invoices_logger.exception(e)
        messages.error(request, str(e), error_extra_tags)
    if generated_invoices:
        messages.success(request, f"{len(generated_invoices)} invoice(s) were generated successfully.", extra_tags)
    else:
        messages.warning(
            request,
            f"No invoices were generated. {'Either the invoices already exist or there was no usage for that month' if not has_errors else ' Please correct the issues and try again.'}",
            warning_extra_tags,
        )
    return redirect("invoices")


@accounting_or_user_office_or_manager_required
@require_GET
def search_invoices(request):
    return queryset_search_filter(
        Invoice.objects.all(),
        ["invoice_number", "project_details__project__name", "project_details__project__account__name"],
        request,
    )


@accounting_or_user_office_or_manager_required
@require_GET
def view_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    dictionary = {
        "invoice": invoice,
        "core_facilities": CoreFacility.objects.exists(),
        "tool_title": category_name_for_item_type(BillableItemType.TOOL_USAGE),
        "area_title": category_name_for_item_type(BillableItemType.AREA_ACCESS),
        "staff_charge_title": category_name_for_item_type(BillableItemType.STAFF_CHARGE),
        "consumable_title": category_name_for_item_type(BillableItemType.CONSUMABLE),
        "training_title": category_name_for_item_type(BillableItemType.TRAINING),
        "missed_reservation_title": category_name_for_item_type(BillableItemType.MISSED_RESERVATION),
        "custom_charge_title": category_name_for_item_type(BillableItemType.CUSTOM_CHARGE),
    }
    invoice_list_ids = (
        request.session["current_invoice_list_id"] if "current_invoice_list_id" in request.session else []
    )
    if invoice.id in invoice_list_ids:
        index = invoice_list_ids.index(invoice.id)
        dictionary.update(
            {
                "previous_id": invoice_list_ids[index - 1] if index - 1 >= 0 else None,
                "next_id": invoice_list_ids[index + 1] if index + 1 < len(invoice_list_ids) else None,
            }
        )

    return render(request, "invoices/invoice.html", dictionary)


def create_rate_link(rate_type, tool, area, consumable):
    try:
        if rate_type and rate_type.type == RateType.Type.CONSUMABLE:
            return reverse("create_rate", args=[RateType.Type.CONSUMABLE, consumable.id])
        elif rate_type and rate_type.get_rate_group_type() == RateType.Type.TOOL:
            return reverse("create_rate", args=[RateType.Type.TOOL, tool.id])
        elif rate_type and rate_type.get_rate_group_type() == RateType.Type.AREA:
            return reverse("create_rate", args=[RateType.Type.AREA, area.id])
        elif rate_type:
            reverse("create_rate", args=[rate_type.type])
    except:
        pass
    return reverse("create_rate")


@accounting_or_manager_required
@require_POST
def review_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    if not invoice.reviewed_date:
        invoice.reviewed_date = timezone.now()
        invoice.reviewed_by = request.user
        invoice.save()
        messages.success(request, f"Invoice {invoice.invoice_number} was successfully marked as reviewed.")
    else:
        messages.error(request, f"Invoice {invoice.invoice_number} has already been reviewed.")
    return redirect("view_invoice", invoice_id=invoice_id)


@accounting_or_manager_required
@require_POST
def send_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    if invoice.reviewed_date:
        if not invoice.project_details.email_to():
            link = reverse("project", args=[invoice.project_details.project.id])
            messages.error(
                request,
                mark_safe(
                    f"Invoice {invoice.invoice_number} could not sent because no email is set on the project - click <a href='{link}'>here</a> to add some"
                ),
            )
        else:
            sent = invoice.send()
            if sent:
                messages.success(request, f"Invoice {invoice.invoice_number} was successfully sent.")
            else:
                messages.error(request, f"Invoice {invoice.invoice_number} could not be sent.")
    else:
        messages.error(request, f"Invoice {invoice.invoice_number} needs to be reviewed before sending.")
    return redirect("view_invoice", invoice_id=invoice_id)


@accounting_or_manager_required
@require_POST
def void_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    processor.void_invoice(invoice, request)
    return redirect("view_invoice", invoice_id=invoice_id)


@accounting_or_manager_required
@require_POST
def delete_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    processor.delete_invoice(invoice, request)
    return redirect("invoices")


@accounting_or_user_office_or_manager_required
@require_POST
def zip_invoices(request, file_type="file"):
    invoice_ids: List[str] = request.POST.getlist("selected_invoice_id[]")
    if not invoice_ids:
        return redirect("invoices")
    else:
        return zip_response(request, Invoice.objects.filter(id__in=invoice_ids), file_type)


@accounting_or_user_office_or_manager_required
@require_POST
def review_and_send_invoices(request):
    invoice_ids: List[str] = request.POST.getlist("selected_invoice_id[]")
    for invoice_id in invoice_ids:
        review_invoice(request, invoice_id)
        send_invoice(request, invoice_id)
    return redirect("invoices")


@accounting_or_user_office_or_manager_required
@require_POST
def mark_invoices_paid_in_full(request):
    invoice_ids: List[str] = request.POST.getlist("selected_invoice_id[]")
    for invoice_id in invoice_ids:
        mark_invoice_paid_in_full(request, invoice_id)
    return redirect("invoices")


@accounting_or_user_office_or_manager_required
@require_http_methods(["GET", "POST"])
def mark_invoice_paid_in_full(request, invoice_id):
    invoice = get_object_or_404(Invoice, pk=invoice_id)
    invoice.mark_as_paid_in_full(request.user)
    return redirect("view_invoice", invoice_id=invoice_id)


@accounting_or_user_office_or_manager_required
@require_GET
def re_render_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, pk=invoice_id)
    try:
        if invoice.file:
            messages.warning(request, "This invoice already has a rendered invoice on file")
        else:
            invoice.render_and_save_file()
            messages.success(request, "Invoice generated successfully")
    except InvoiceGenerationException as e:
        extra_tags = "data-speed=30000"
        link = reverse("re_render_invoice", args=[e.invoice.id])
        message = e.msg + f" - click <a href='{link}'>here</a> to try generating it again."
        messages.error(request, mark_safe(message), extra_tags)
    return redirect(request.META.get("HTTP_REFERER", "invoices"))


@accounting_or_user_office_or_manager_required
@require_GET
def csv_invoice(request, invoice_id):
    invoice = get_object_or_404(Invoice, pk=invoice_id)
    try:
        invoice_renderer = CSVInvoiceRenderer()
        content = invoice_renderer.render_invoice(invoice)
        content.seek(0)
        response = HttpResponse(content.read(), content_type="text/csv")
        response["Content-Disposition"] = (
            f'attachment; filename="{invoice.filename(invoice_renderer.get_file_extension())}"'
        )
        return response
    except InvoiceGenerationException as e:
        messages.error(request, e.msg)
        return HttpResponse()


@accounting_or_manager_required
@require_POST
def invoice_payment_received(request, invoice_id):
    invoice = get_object_or_404(Invoice, id=invoice_id)
    amount = Decimal(request.POST["payment_received_amount"])
    note = request.POST.get("payment_note")
    received = datetime.strptime(request.POST["payment_received_date"], date_input_format)
    payment = invoice.record_payment(request.user, amount, received, note=note)
    messages.success(
        request,
        f"The payment of {payment.amount_display()} for invoice {invoice.invoice_number} was marked as received on {date_format(payment.payment_received)}.",
    )
    return redirect("view_invoice", invoice_id=invoice_id)


@accounting_or_manager_required
@require_POST
def invoice_payment_processed(request, payment_id):
    payment = get_object_or_404(InvoicePayment, id=payment_id)
    payment.updated_by = request.user
    payment.payment_processed = datetime.strptime(request.POST["payment_processed_date"], date_input_format)
    payment.save()
    messages.success(
        request,
        f"The payment of {payment.amount_display()} for invoice {payment.invoice.invoice_number} was marked as processed on {date_format(payment.payment_processed)}.",
    )
    return redirect("view_invoice", invoice_id=payment.invoice_id)


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def send_invoice_payment_reminder(request):
    return do_send_invoice_payment_reminder()


def do_send_invoice_payment_reminder():
    today = timezone.now()
    unpaid_invoices = Invoice.objects.filter(due_date__lte=today, voided_date=None)
    for unpaid_invoice in unpaid_invoices:
        if unpaid_invoice.total_outstanding_amount() > Decimal(0):
            if not unpaid_invoice.last_reminder_sent_date:
                unpaid_invoice.send_reminder()
            else:
                # Check days since last reminder sent
                time_diff = today - unpaid_invoice.last_reminder_sent_date
                too_long_since_last = (
                    unpaid_invoice.configuration.reminder_frequency
                    and time_diff.days >= unpaid_invoice.configuration.reminder_frequency
                )
                # Send reminder if none has been sent yet, or if it's been too long
                if too_long_since_last:
                    unpaid_invoice.send_reminder()
    return HttpResponse()


def zip_response(request, invoice_list: List[Invoice], file_type="file"):
    generated_date = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    parent_folder_name = f"invoices_{generated_date}"
    zip_io = io.BytesIO()
    with zipfile.ZipFile(zip_io, mode="w", compression=zipfile.ZIP_DEFLATED) as backup_zip:
        csv_invoice_renderer = CSVInvoiceRenderer()
        for invoice in invoice_list:
            if file_type.lower() == "csv":
                ext = csv_invoice_renderer.get_file_extension()
                content = csv_invoice_renderer.render_invoice(invoice)
                content.seek(0)
                backup_zip.writestr(f"{parent_folder_name}/" + invoice.filename_for_zip(ext), content.read())
                content.close()
            elif invoice.file:
                backup_zip.write(invoice.file.path, f"{parent_folder_name}/" + invoice.filename_for_zip())
    response = HttpResponse(zip_io.getvalue(), content_type="application/x-zip-compressed")
    response["Content-Disposition"] = "attachment; filename=%s" % parent_folder_name + ".zip"
    response["Content-Length"] = zip_io.tell()
    return response



# Generated by Django 3.2.19 on 2023-05-31 17:04
from copy import deepcopy
from typing import Dict

from django.db import DatabaseError, migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor


class Migration(migrations.Migration):

    dependencies = [
        ("cap_discount", "0003_multi_tier_discount"),
    ]

    operations = []

    def apply(self, project_state, schema_editor: BaseDatabaseSchemaEditor, collect_sql=False):
        # Because we changed the initial migration, we need to add/remove the constraints manually
        # We check if the old constraints exist in the DB (case when app was migrated before the change)
        # and remove them if they do. Then we add the new constraints
        current_name_of_constraint = "capdiscountconfiguration_unique_rate_category_facility"
        current_name_of_null_constraint = "capdiscountconfiguration_unique_rate_category_null_facility"
        model_state = project_state.models["cap_discount", "capdiscountconfiguration"]
        unique_constr_definition = model_state.get_constraint_by_name(current_name_of_constraint)
        unique_null_constr_definition = model_state.get_constraint_by_name(current_name_of_null_constraint)
        previous_unique_constr = deepcopy(unique_constr_definition)
        previous_unique_null_constr = deepcopy(unique_null_constr_definition)
        CapDiscountConfiguration = project_state.apps.get_model("cap_discount", "CapDiscountConfiguration")
        constraints: Dict[str, Dict] = schema_editor.connection.introspection.get_constraints(
            schema_editor.connection.cursor(), "cap_discount_capdiscountconfiguration"
        )
        # Here we are just going to remove the unique constraints and re-add them
        db_previous_constr_names = []
        db_previous_null_constr_names = []
        for constraint_name, constraint in constraints.items():
            if "capdiscountconfiguration_unique_rate" in constraint_name and constraint.get("unique") == True:
                if "_null" in constraint_name:
                    db_previous_null_constr_names.append(constraint_name)
                else:
                    db_previous_constr_names.append(constraint_name)
        for db_previous_constr_name in db_previous_constr_names:
            previous_unique_constr.name = db_previous_constr_name
            try:
                schema_editor.remove_constraint(CapDiscountConfiguration, previous_unique_constr)
            except DatabaseError:
                pass
        for db_previous_null_constr_name in db_previous_null_constr_names:
            previous_unique_null_constr.name = db_previous_null_constr_name
            try:
                schema_editor.remove_constraint(CapDiscountConfiguration, previous_unique_null_constr)
            except DatabaseError:
                pass
        try:
            schema_editor.add_constraint(CapDiscountConfiguration, unique_constr_definition)
        except DatabaseError:
            pass
        try:
            schema_editor.add_constraint(CapDiscountConfiguration, unique_null_constr_definition)
        except DatabaseError:
            pass
        return project_state
